import bs
import bsGame
import bsMap

class TutorialActivity(bs.Activity):

    def __init__(self,session,settings={}):
        bs.Activity.__init__(self,session,settings)
        self._mapType = bsMap.RampageMap
        self._mapType.preload()

        self._jumpButtonTex = bs.getTexture('buttonJump')
        self._pickUpButtonTex = bs.getTexture('buttonPickUp')
        self._bombButtonTex = bs.getTexture('buttonBomb')
        self._punchButtonTex = bs.getTexture('buttonPunch')
        self._R = bs.getResource('tutorial')
        
    def onTransitionIn(self):
        bs.Activity.onTransitionIn(self)
        self._map = self._mapType(self,bs.Actor.UNRETAINED)

    def onBegin(self):
        bs.Activity.onBegin(self)
        #bs.gameTimer(5000,bs.WeakCall(self.end))

        pos = (0,5.5,-4.4)
        
        s = self._spaz1 = bs.Spaz(self,bs.Actor.UNRETAINED)
        s.handleMessage(bs.StandMessage(pos,0))


        # need different versions of this: taps/buttons/keys
        t = self._tempText = bs.newNode('text',
                                        owner=self,
                                        attrs={'text':'UNDER-CONSTRUCTION\n- this will be a short tutorial -',
                                               'scale':[1.5],
                                               'translate':(0,150),
                                               'hAlign':'center',
                                               'vAlign':'center',
                                               'vAttach':'center'})
        bs.animate(t,'opacity',{0:0.0,1000:0.7})
        

        # need different versions of this: taps/buttons/keys
        t = self._skipText = bs.newNode('text',
                                        owner=self,
                                        attrs={'text':self._R.toSkipPressAnythingText,
                                               'scale':[1.1],
                                               'translate':(0,70),
                                               'hAlign':'center',
                                               'vAlign':'center',
                                               'vAttach':'bottom'})
        bs.animate(t,'opacity',{1000:0.0,2000:0.7})
        t = self._skipCountText = bs.newNode('text',
                                                 owner=self,
                                                 attrs={'text':'',
                                                        'scale':[1.4],
                                                        'translate':(0,30),
                                                        'hAlign':'center',
                                                        'vAlign':'center',
                                                        'vAttach':'bottom'})

        scale = 1.0
        offs = 74.0*scale
        position=(390,120)
        imageSize = 90.0*scale
        p = (position[0],position[1]-offs)
        c = (0.4,1,0.4)
        # self._jumpImage = bs.newNode('image',owner=self,
        #                        attrs={'texture':self._jumpButtonTex,'absoluteScale':True,
        #                                'translate':p,'scale':(imageSize,imageSize),'color':c})
        
        # lets check this once per second just in case..
        # self._skipTimer = bs.Timer(1000,bs.WeakCall(self._updateSkipVotes),repeat=True)
        
    def _updateSkipVotes(self):
        count = sum(1 for player in self.players if player.gameData['pressed'])
        #self._skipCountText.text = (str(total)+'/'+str(len(self.players))+' skip votes') if total > 0 else ''
        self._skipCountText.text = self._R.skipVoteCountText.replace('${COUNT}',str(count)).replace('${TOTAL}',str(len(self.players))) if count > 0 else ''
        if count >= len(self.players):
            bs.playSound(bs.getSound('swish'))
            self._skipCountText.text = self._R.skippingText
            self._skipText.text = ''
            self.end()
            

    def _playerPressedButton(self,player):
        player.gameData['pressed'] = True
        self._updateSkipVotes()
        
    def onPlayerJoined(self,player):
        bs.Activity.onPlayerJoined(self,player)
        player.gameData['pressed'] = False
        # we just wanna know if this player presses anything..
        player.assignInput(('jumpPress','punchPress','bombPress','pickUpPress'),bs.WeakCall(self._playerPressedButton,player))

    def onPlayerLeaving(self,player):
        bs.Activity.onPlayerLeaving(self,player)
        # our leaving may influence the vote total needed/etc
        self._updateSkipVotes()
        
        # if all our players leave we wanna quit out of the session
        # (otherwise we might lead into a co-op game with zero players or something)
        if len(self.players) == 0:
            self.getSession().end()

    def __del__(self):
        bs.Activity.__del__(self)
