import bs
import weakref
import os
import thread
import threading
import time
import random
import urllib
import urllib2
import ast
import gc
import types
import copy
import bsInternal
import bsGame

gConfigFileIsHealthy = False
gCloudConfig = {}

# config keys that we attempt to store in the cloud to share between devices
gCloudConfigKeys = ['launchCount','Campaigns','Player Profiles','Submit Debug Info','Default Player Profiles',
                    'Team Tournament Playlists','Team Tournament Playlist Randomize','Team Tournament Playlist Selection',
                    'Free-for-All Playlists','Free-for-All Playlist Randomize','Free-for-All Playlist Selection',
                    'Selected Coop Game','Show Player Names','Kick Idle Players']


# we sync achievements to/from the cloud only if we have no game-service... Otherwise that's the game-service's job
if bsInternal._getGameService() is None:
    gCloudConfigKeys.append('Achievements')

_gPlayerColors = [(1,0.25,0.2),(0.2,1,0.2),(0.1,0.25,1),(0.2,1,1),
                  (0.65,0.1,1.0),(1,1,0),(1,0.5,0),(1,0.3,0.5),
                  (0.1,0.1,0.5),(0.4,0.2,0.1),(0.1,0.35,0.1),(1,0.8,0.5),
                  (0.4,0.05,0.05),(0.13,0.13,0.13),(0.5,0.5,0.5),(1,1,1)]
gMusic = None

gNodeOwnerWeakRefs = []
gNodeOwnerWeakRefsCleanCounter = 0

gPrintOnceErrors = set()

def printErrorOnce(e,traceback=True):
    """
    category: General Utility Functions

    Given an error string, prints it along with (optionally) a traceback.
    A given error string is only printed once; the same error string submitted
    again will be ignored.

    This is handy for catching and reporting non-fatal errors without risking
    overwhelming the log.
    """
    if not e in gPrintOnceErrors:
        gPrintOnceErrors.add(e)
        print 'ERROR: '+e
        if traceback:
            import traceback
            traceback.print_stack()
        
def getGlobals():
    """
    category: Game Flow Functions
    
    Returns the globals node, containing various things like screen color tints
    and camera controls.
    """
    
    # fixme - we shoudln't be shoving globals into the bs module as an attribute..
    return bs.globals

def getHumanReadableUserScriptsPath():
    "Return a human readable path to user-scripts (NOT a valid filesystem path)"
    env = bs.getEnvironment()
    path = env['userScriptsDirectory']
    if path is None: return '<Not Available>'
    # on newer versions of android, the user's external storage dir is probably
    # only visible to the user's processes and thus not really useful printed in its entirety
    # lets print it as <External Storage>/myfilepath
    if env['platform'] == 'android':
        extStoragePath = bsInternal._androidGetExternalStoragePath()
        if extStoragePath is not None and env['userScriptsDirectory'].startswith(extStoragePath):
            path = '<'+bs.getResource('externalStorageText')+'>'+env['userScriptsDirectory'][len(extStoragePath):]
    return path
    
def showUserScripts():
    "Open or at least nicely print the location of the user-scripts directory"
    env = bs.getEnvironment()
    if 'Mac' in env['userAgentString']: os.system("open '"+bs.getEnvironment()['userScriptsDirectory']+"'")
    else:
        bs.screenMessage(getHumanReadableUserScriptsPath())

def createUserSystemScripts():
    'Sets up a copy of BombSquad system scripts under your user scripts dir that you can edit and experiment with.'
    import shutil
    path = bs.getEnvironment()['userScriptsDirectory']+'/sys/'+bs.getEnvironment()['version']
    if os.path.exists(path): shutil.rmtree(path)
    if os.path.exists(path+"_tmp"): shutil.rmtree(path+"_tmp")

    try:
        os.makedirs(path+'_tmp')
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path+"_tmp"): pass
        else: raise

    # shutil.copytree doesn't seem to work nicely on android, so lets do it manually
    srcDir = bs.getEnvironment()['systemScriptsDirectory']
    dstDir = path+"_tmp"
    files = os.listdir(bs.getEnvironment()['systemScriptsDirectory'])
    for f in files:
        print 'COPYING',srcDir+'/'+f,'->',dstDir
        shutil.copyfile(srcDir+'/'+f,dstDir+'/'+f)

    print 'MOVING',path+"_tmp",path
    shutil.move(path+"_tmp",path)
    #if 'android' in bs.getEnvironment()['userAgentString']: bs.androidRefreshFiles() # get android file transfer to reflect this
    print 'Created system scripts at :\''+path+'\'\nRestart BombSquad to use them. (use bs.quit() to exit the game)'
    if bs.getEnvironment()['platform'] == 'android':
        print 'Note: the new files may not be visible via android-file-transfer until you restart your device.'

def deleteUserSystemScripts():
    'Cleans out the scripts created by createUserSystemScripts()'
    import shutil
    path = bs.getEnvironment()['userScriptsDirectory']+'/sys/'+bs.getEnvironment()['version']
    if os.path.exists(path):
        shutil.rmtree(path)
        print 'User system scripts deleted.\nRestart BombSquad to use internal scripts. (use bs.quit() to exit the game)'
    else:
        print 'User system scripts not found.'

    # if the sys path is empty, kill it
    d = bs.getEnvironment()['userScriptsDirectory']+'/sys'
    if os.path.isdir(d) and len(os.listdir(d)) == 0:
        os.rmdir(d)
    if 'android' in bs.getEnvironment()['userAgentString']: bs.androidRefreshFiles() # get android file transfer to reflect this


class WeakCall(object):
    """
    category: General Utility Classes

    Wraps a callable and arguments into a single callable object.
    When passed a bound method as the callable, the 'self' portion
    of it is weak-referenced, meaning the underlying instance is
    free to die if not referenced from elsewhere. Should this
    occur, calling the WeakCall is simply a no-op.

    Think of this simply as a way to tell an object to do something
    at some point in the future if it happens to still be alive.
    """
    def __init__(self,*args,**keywds):
        """
        Instantiate a WeakCall; pass a callable as the first
        arg, followed by any number of arguments or keywords.

        # example: wrap a method call with some positional and keyword args:
        myCallable = bs.WeakCall(myObj.doStuff,argVal,namedArg=anotherVal)

        # now we have a single callable to run that whole mess
        myCallable()
        """
        if hasattr(args[0],'im_func'): self._call = WeakMethod(args[0])
        else:
            global _gDidWeakCallWarning
            try: did = _gDidWeakCallWarning
            except Exception: did = False
            if not did:
                print 'Warning: callable passed to bs.WeakCall() is not weak-referencable ('+str(args[0])+'); use bs.Call() instead to avoid this warning. Stack-trace:'
                import traceback
                traceback.print_stack()
                _gDidWeakCallWarning = True
            self._call = args[0]
        self._args = args[1:]
        self._keywds = keywds
    def __call__(self,*argsExtra):
        return self._call(*self._args+argsExtra,**self._keywds)

class Call(object):
    """
    category: General Utility Classes

    Wraps a callable and arguments into a single callable object.
    The callable is strong-referenced so it won't die until this object does.
    Note that a bound method (ex: myObj.doSomething) contains a reference
    to the 'self' object, so you will be keeping that object alive too.
    Use bs.WeakCall if you want to pass a callback without keeping things alive.
    """
    def __init__(self,*args,**keywds):
        """
        Instantiate a Call; pass a callable as the first
        arg, followed by any number of arguments or keywords.

        # example: wrap a method call with 1 positional and 1 keyword arg
        myCallable = bs.Call(myObj.doStuff,argVal,namedArg=anotherVal)

        # now we have a single callable to run that whole mess
        myCallable()
        """

        self._call = args[0]
        self._args = args[1:]
        self._keywds = keywds
    def __call__(self,*argsExtra):
        return self._call(*self._args+argsExtra,**self._keywds)

class WeakMethod :
    """ wraps a bound method using weak references so that the original is free to die.
    If called with a dead target, is simply a no-op """
    def __init__( self , f ) :
        self.f = f.im_func
        self.c = weakref.ref( f.im_self )
    def __call__( self , *args, **keywds ) :
        c = self.c()
        if c is None : return
        return self.f(*((c,)+args),**keywds)

    
_gLanguage = None

def _getLanguages():
    langs = set()
    env = bs.getEnvironment()
    for i,d in enumerate([env['systemScriptsDirectory'],env['userScriptsDirectory']]):
        for name in os.listdir(d):
            if name.startswith('bsLanguage') and name.endswith('.py'):
                langs.add(name[10:-3])
    vals = list(langs)
    vals.sort()
    return vals

# used internally
def _handleRemoteAchievementList(completedAchievements):
    import bsAchievement
    bsAchievement.setCompletedAchievements(completedAchievements)

# used internally
def _handleRemoteConfigChange(config,overwrite):

    wroteSomething = False

    global gCloudConfig

    # if we receieved cloud config data that includes a launch-count substantially less than ours, lets
    # ignore the whole thing.  I've seen cases where initial cloud values never come in on
    # the first launch of an app, and this hopefully prevents that app from overwriting
    # everyone else's config when it spits its default values out to the cloud

    if overwrite and 'launchCount' in config and 'launchCount' in bs.config:
        # we listen to syncs that are witin one of our launch-count
        if config['launchCount'] < bs.config['launchCount'] - 1:
            return

    for c in config.items():

        try:
            # we store what we know the remote state to be to avoid sending redundant sets
            gCloudConfig[c[0]] = c[1]

            # eval the actual value from the string
            val = ast.literal_eval(c[1])

            # we commit the value to our local config only if overwrite is true or it doesn't exist
            # (and its a key we expect to find in the cloud)
            if c[0] in gCloudConfigKeys and (overwrite or (c[0] not in bs.config)):

                wroteSomething = True
                # do a bit of filtering here; for some values (such as, say, launchCount)
                # we want to use whatever's bigger and not just whichever's newer
                if c[0] == 'launchCount' and 'launchCount' in bs.config:
                    bs.config['launchCount'] = max(bs.config['launchCount'],val)

                # for any other vals we just overwrite with the newer cloud version
                else:
                    bs.config[c[0]] = val

        except Exception,e:
            print 'Exception setting remote config value for',c[0],'to',c[1],':',e

    # commit this stuff locally..
    if wroteSomething:
        writeConfig(storeRemote=False)

def _shouldSubmitDebugInfo():
    try:
        if bs.getEnvironment()['testBuild']: return True
        else: return bs.config['Submit Debug Info']
    except Exception: return True

def _handleBecomingActive():
    # if there's music playing externally, make sure we aren't playing ours
    import bsUtils
    import bsInternal
    if bsInternal._isOSPlayingMusic():
        playMusic(None)

# high level way to launch a challenge game locally
def _handleRunChallengeGame(game):
    import bsCoopGame
    import bsUI
    if game == '': raise Exception("empty game name")
    campaignName,levelName = game.split(':')
    campaign = bsCoopGame.getCampaign(campaignName)
    levels = campaign.getLevels()
    # if this campaign is sequential, make sure we've completed the one before this
    if campaign.isSequential():
        for i in range(len(levels)):
            if levels[i]['name'] == levelName:
                break
            if not levels[i]['Complete']:
                bsUI.LockedErrorWindow(bs.translate('coopLevelNames',levelName),bs.translate('coopLevelNames',levels[i]['name']))
                return False

    # if we're an expired demo, bring up the purchase ui for convenience
    if bsInternal._getDemoTimeRemaining() == 0:
        bsUI.showDemoExpiredUI()
        return

    # ok we're good to go..
    bsUI.gCampaignSelection = {'campaign':campaignName,'level':levelName}
    def _fadeEnd():
        bsInternal._hostSession('bsCoopGame.CoopSession')
    bsInternal._fadeLocalScreen(False,time=250,endCall=_fadeEnd)
    return True

def openURL(address):
    """
    category: General Utility Functions

    Open the provided url in a web-browser, or display the URL
    string in a window if that isn't possible.
    """
    bsInternal._openURL(address)

def _getDefaultFreeForAllPlaylist():
    return [
         {
            'map':'Doom Shroom',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':10,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'map':'Monkey Face',
            'settings':{
               'Chosen One Gets Gloves':True,
               'Chosen One Gets Shield':False,
               'Chosen One Time':30,
               'Epic Mode':0,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsChosenOne.ChosenOneGame'
         },
         {
            'map':'Zigzag',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKingOfTheHill.KingOfTheHillGame'
         },
         {
            'map':'Tip Top',
            'settings':{
               'Epic Mode':1,
               'Lives Per Player':1,
               'Respawn Times':1.0,
               'Time Limit':120
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'map':'The Pad',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKeepAway.KeepAwayGame'
         },
         {
            'map':'Rampage',
            'settings':{
               'Epic Mode':True,
               'Kills to Win Per Player':10,
               'Respawn Times':0.25,
               'Time Limit':120
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'map':'Big G',
            'settings':{
               'Epic Mode':False,
               'Laps':3,
               'Mine Spawn Interval':4000,
               'Mine Spawning':4000,
               'Time Limit':300
            },
            'type':'bsRace.RaceGame'
         },
         {
            'map':'Happy Thoughts',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKingOfTheHill.KingOfTheHillGame'
         },
         {
            'map':'Step Right Up',
            'settings':{
               'Epic Mode':False,
               'Lives Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'map':'Crag Castle',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':10,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         }
      ]

def _getDefaultTeamsPlaylist():
    return [
         {
            'map':'Bridgit',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':0,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':600
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'map':'Step Right Up',
            'settings':{
               'Epic Mode':False,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':600
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'map':'Rampage',
            'settings':{
               'Balance Total Lives':False,
               'Epic Mode':False,
               'Lives Per Player':3,
               'Respawn Times':1.0,
               'Solo Mode':True,
               'Time Limit':600
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'map':'Roundabout',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'map':'Hockey Stadium',
            'settings':{
               'Respawn Times':1.0,
               'Score to Win':1,
               'Time Limit':600
            },
            'type':'bsHockey.HockeyGame'
         },
         {
            'map':'Monkey Face',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKeepAway.KeepAwayGame'
         },
         {
            'map':'Tip Top',
            'settings':{
               'Balance Total Lives':False,
               'Epic Mode':True,
               'Lives Per Player':1,
               'Respawn Times':1.0,
               'Solo Mode':False,
               'Time Limit':120
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'map':'Crag Castle',
            'settings':{
               'Epic Mode':False,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':300
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'map':'Doom Shroom',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'map':'Roundabout',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':0,
               'Respawn Times':1.0,
               'Score to Win':2,
               'Time Limit':600
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'map':'Football Stadium',
            'settings':{
               'Respawn Times':1.0,
               'Score to Win':21,
               'Time Limit':600
            },
            'type':'bsFootball.FootballTeamGame'
         },
         {
            'map':'Bridgit',
            'settings':{
               'Epic Mode':True,
               'Respawn Times':0.25,
               'Score to Win':3,
               'Time Limit':120
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'map':'Tip Top',
            'settings':{
               'Hold Time':30,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsKingOfTheHill.KingOfTheHillGame'
         },
         {
            'map':'Zigzag',
            'settings':{
               'Epic Mode':False,
               'Respawn Times':1.0,
               'Score to Win':2,
               'Time Limit':300
            },
            'type':'bsAssault.AssaultGame'
         },
         {
            'map':'Happy Thoughts',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':0,
               'Respawn Times':1.0,
               'Score to Win':3,
               'Time Limit':300
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'map':'Big G',
            'settings':{
               'Epic Mode':True,
               'Laps':1,
               'Mine Spawning':2000,
               'Time Limit':300
            },
            'type':'bsRace.RaceGame'
         },
         {
            'map':'Monkey Face',
            'settings':{
               'Epic Mode':False,
               'Kills to Win Per Player':5,
               'Respawn Times':1.0,
               'Time Limit':300
            },
            'type':'bsDeathMatch.DeathMatchGame'
         },
         {
            'map':'Tip Top',
            'settings':{
               'Epic Mode':False,
               'Flag Idle Return Time':30,
               'Flag Touch Return Time':3,
               'Respawn Times':1.0,
               'Score to Win':2,
               'Time Limit':300
            },
            'type':'bsCaptureTheFlag.CTFGame'
         },
         {
            'map':'Crag Castle',
            'settings':{
               'Balance Total Lives':False,
               'Epic Mode':False,
               'Lives Per Player':3,
               'Respawn Times':1.0,
               'Solo Mode':False,
               'Time Limit':300
            },
            'type':'bsElimination.EliminationGame'
         },
         {
            'map':'Zigzag',
            'settings':{
               'Epic Mode':True,
               'Respawn Times':0.25,
               'Time Limit':120
            },
            'type':'bsConquest.ConquestGame'
         }
      ]

def _readConfig():
    global gConfigFileIsHealthy
    configFilePath = bs.getEnvironment()['configFilePath']
    configContents = ''
    try:
        if os.path.exists(configFilePath):
            f = open(configFilePath,'rb')
            configContents = f.read()
            f.close()
            bs.config = eval(configContents)
        else: bs.config = {}
        gConfigFileIsHealthy = True
    except Exception,e:
        print 'error reading config file at time '+str(bs.getRealTime())+': \''+configFilePath+'\':\n',e
        # whenever this happens lets back up the broken one just in case it gets overwritten accidentally
        print 'backing up current config file to \''+configFilePath+".broken\'"
        try:
            import shutil
            shutil.copyfile(configFilePath,configFilePath+'.broken')
        except Exception,e: print 'EXC copying broken config:',e
        try: bsInternal._log('broken config contents:\n'+configContents.replace('\000','<NULL_BYTE>')+'\n',toConsole=False)
        except Exception,e: print 'EXC logging broken config contents:',e
        bs.config = {}

        # now attempt to read one of our 'prev' backup copies
        prevPath = configFilePath + '.prev'
        try:
            if os.path.exists(prevPath):
                f = open(prevPath,'rb')
                configContents = f.read()
                f.close()
                bs.config = eval(configContents)
            else: bs.config = {}
            gConfigFileIsHealthy = True
            print 'successfully read backup config.'
        except Exception,e:
            print 'EXC reading prev backup config:',e
            
    global gCloudConfigKeys
    bsInternal._requestRemoteConfigKeys(gCloudConfigKeys)


def _prettifyList(l,indent=0,multiLine=True):
    bits = []
    bits.append('[\n' if multiLine else '[')
    for i in range(len(l)):
        a = l[i]
        if isinstance(a,dict):
            if multiLine: bits.append(' '*(indent+3))
            bits += _prettifyDict(a,indent+3,multiLine)
        else:
            if multiLine: bits.append(' '*(indent+3) + repr(a))
            else: bits.append(repr(a))
        if (i < (len(l)-1)):
            bits.append(',')
        if multiLine: bits.append("\n")
    bits.append((' '*indent + ']') if multiLine else ']')
    return bits

def _prettifyDict(d,indent=0,multiLine=True):
    bits = []
    bits.append('{\n' if multiLine else '{')
    items = d.items()
    items.sort()
    for i in range(len(items)):
        a = items[i][0]
        b = items[i][1]
        if isinstance(b,dict):
            if multiLine: bits.append(' '*(indent+3) + repr(a) + ':')
            else: bits.append(repr(a) + ':')
            bits += _prettifyDict(b,indent+3,multiLine)
        elif isinstance(b,list):
            if multiLine: bits.append(' '*(indent+3) + repr(a) + ':')
            else: bits.append(repr(a) + ':')
            bits += _prettifyList(b,indent+3,multiLine)
        else:
            if multiLine: bits.append(' '*(indent+3) + repr(a) + ':' + repr(b))
            else: bits.append(repr(a) + ':' + repr(b))
        if (i < (len(items)-1)):
            bits.append(',')
        if multiLine: bits.append("\n")
    bits.append((' '*indent + '}') if multiLine else '}')
    return bits


def _getAllGameTypes():
    allGames = []
    haveMods = False
    env = bs.getEnvironment()
    for i,d in enumerate([env['systemScriptsDirectory'],env['userScriptsDirectory']]):
        for name in os.listdir(d):
            try:
                if name == 'sys' or name.endswith('.py'):
                    # if there's anything in user-mods, let the game know that we've got mods going on
                    # (we do less error reporting in this case for my sanity's sake)
                    if i == 1: haveMods = True
                    
                if name.endswith('.py') and not name.startswith('bsLanguage') and not name in ['bsMeteorShowerInternal.py']:
                    module = __import__(name.replace('.py',''))
                    if hasattr(module,'getGames'):
                        games = module.getGames()
                        allGames += games
            except Exception,e:
                print 'Error importing game module \''+name+'\':',e

    bsInternal._setHaveMods(haveMods)
    return allGames



gTips = []

def _getNextTip():
   """ returns the next tip to be displayed """
   global gTips
   if len(gTips) == 0:
       for tip in _getAllTips():
           gTips.insert(random.randint(0,len(gTips)),tip)
   return gTips.pop()

def _getAllTips():
    tips =  [
        'If you are short on controllers, install the \'BombSquad Remote\' app\non your iOS or Android devices to use them as controllers.',
        'Create player profiles for yourself and your friends with\nyour preferred names and appearances instead of using random ones.',
        'You can \'aim\' your punches by spinning left or right.\nThis is useful for knocking bad guys off edges or scoring in hockey.',
        'If you pick up a curse, your only hope for survival is to\nfind a health powerup in the next few seconds.',
        'A perfectly timed running-jumping-spin-punch can kill in a single hit\nand earn you lifelong respect from your friends.',
        'Always remember to floss.',
        'Don\'t run all the time.  Really.  You will fall off cliffs.',
        'In Capture-the-Flag, your own flag must be at your base to score, If the other\nteam is about to score, stealing their flag can be a good way to stop them.',
        'If you get a sticky-bomb stuck to you, jump around and spin in circles. You might\nshake the bomb off, or if nothing else your last moments will be entertaining.',
        'You don\'t need to edit your profile to change characters; Just press the top\nbutton (pick-up) when joining a game to override your default.',
        'You take damage when you whack your head on things,\nso try to not whack your head on things.',
        'If you kill an enemy in one hit you get double points for it.',
        'Despite their looks, all characters\' abilities are identical,\nso just pick whichever one you most closely resemble.',
        'You can throw bombs higher if you jump just before throwing.',
        'Throw strength is based on the direction you are holding.\nTo toss something gently in front of you, don\'t hold any direction.',
        'If someone picks you up, punch them and they\'ll let go.\nThis works in real life too.',
        'Don\'t get too cocky with that energy shield; you can still get yourself thrown off a cliff.',
        'Many things can be picked up and thrown, including other players.  Tossing\nyour enemies off cliffs can be an effective and emotionally fulfilling strategy.',
        'Ice bombs are not very powerful, but they freeze\nwhoever they hit, leaving them vulnerable to shattering.',
        'Run back and forth before throwing a bomb\nto \'whiplash\' it and throw it farther.',
        'Punches do more damage the faster your fists are moving,\nso try running, jumping, and spinning like crazy.',
        'In hockey, you\'ll maintain more speed if you turn gradually.',
        'The head is the most vulnerable area, so a sticky-bomb\nto the noggin usually means game-over.',
        'Hold down any button to run. You\'ll get places faster\nbut won\'t turn very well, so watch out for cliffs.',
        'Powerups only have time limits in co-op games.\nIn teams and free-for-all they\'re yours until you die.',
        'You can judge when a bomb is going to explode based on the\ncolor of sparks from its fuse:  yellow..orange..red..BOOM.',
        ]

    ua = bs.getEnvironment()['userAgentString']

    # graphics settings applies to desktops and android
    if bs.getEnvironment()['interfaceType'] == 'desktop' or ('android' in ua and not 'OUYA' in ua and not bsInternal._isRunningOnFantasia()):
        tips += [
            'If your framerate is choppy, try turning down resolution\nor visuals in the game\'s graphics settings.',
            ]
    ua = bs.getEnvironment()['userAgentString']

    # auto-kick players applies to desktop and ouya
    if bs.getEnvironment()['interfaceType'] == 'desktop' or 'OUYA' in ua:
        tips += [
            'If you\'ve got lots of players coming and going, turn on \'auto-kick-idle-players\'\nunder settings in case anyone forgets to leave the game.',
            ]

    # hot-plugging is currently mac/android only
    if 'Mac' in ua or 'android' in ua:
        tips += [
            'Players can join and leave in the middle of most games,\nand you can also plug and unplug gamepads on the fly.',
        ]

    return tips

def _filterPlaylist(playlist,sessionType,addResolvedType=False):
    """ returns a filtered version of a playlist - strips out or replaces invalid game types,
    makes sure all settings are present, and adds in a 'resolvedType' which is the actual type """
    
    goodList = []
    for i,gamespec in enumerate(playlist):

        # 'map' used to be called 'level'
        if 'level' in gamespec:
            gamespec['map'] = gamespec['level']
            del gamespec['level']

        # ok, for each game in our list, try to import the module and grab the actual game class.
        # add successful ones to our initial list to present to the user
        if type(gamespec['type']) in [str,unicode]:
            try:
                # do some type filters for backwards compat.
                if gamespec['type'] == 'Happy_Thoughts.HappyThoughtsGame': gamespec['type'] = 'bsAssault.AssaultGame'
                if gamespec['type'] == 'Assault.AssaultGame': gamespec['type'] = 'bsAssault.AssaultGame'
                if gamespec['type'] == 'King_of_the_Hill.KingOfTheHillGame': gamespec['type'] = 'bsKingOfTheHill.KingOfTheHillGame'
                if gamespec['type'] == 'Capture_the_Flag.CTFGame': gamespec['type'] = 'bsCaptureTheFlag.CTFGame'
                if gamespec['type'] == 'Death_Match.DeathMatchGame': gamespec['type'] = 'bsDeathMatch.DeathMatchGame'
                if gamespec['type'] == 'ChosenOne.ChosenOneGame': gamespec['type'] = 'bsChosenOne.ChosenOneGame'
                if gamespec['type'] == 'Conquest.Conquest': gamespec['type'] = 'bsConquest.ConquestGame'
                if gamespec['type'] == 'Conquest.ConquestGame': gamespec['type'] = 'bsConquest.ConquestGame'
                if gamespec['type'] == 'Elimination.EliminationGame': gamespec['type'] = 'bsElimination.EliminationGame'
                if gamespec['type'] == 'Football.FootballGame': gamespec['type'] = 'bsFootball.FootballTeamGame'
                if gamespec['type'] == 'Hockey.HockeyGame': gamespec['type'] = 'bsHockey.HockeyGame'
                if gamespec['type'] == 'Keep_Away.KeepAwayGame': gamespec['type'] = 'bsKeepAway.KeepAwayGame'
                if gamespec['type'] == 'Race.RaceGame': gamespec['type'] = 'bsRace.RaceGame'

                if gamespec['map'] == 'AlwaysLand': gamespec['map'] = 'Happy Thoughts'

                gameModuleName,gameClassName = gamespec['type'].split('.')

                gameModule = __import__(gameModuleName)
                gameClass = getattr(gameModule,gameClassName)

                if addResolvedType: gamespec['resolvedType'] = gameClass

                # make sure all settings the game defines are present
                neededSettings = gameClass.getSettings(sessionType)
                for settingName,setting in neededSettings:
                    if settingName not in gamespec['settings']:
                        gamespec['settings'][settingName] = setting['default']

                goodList.append(gamespec)

            except Exception,e:
                print 'Error resolving game \''+gamespec['type']+"':",e
        else:
            raise Exception("invalid gamespec format")

    return goodList


gLogUploadTimerStarted = False
gLogHaveNew = False


# called on debug log prints..
# when this happens, we can upload our log to the server
# after a short bit if desired.
def _handleLog():
    global gLogUploadTimerStarted
    global gLogHaveNew
    gLogHaveNew = True

    if not gLogUploadTimerStarted:
        def _putLog():
            import bsUtils
            if not bsUtils._gSuppressDebugReports:

                try: sessionName = str(bs.getSession())
                except Exception: sessionName = 'unavailable'
                try: activityName = str(bs.getSession().getActivity())
                except Exception: activityName = 'unavailable'

                env = bs.getEnvironment()
                info = {'log':bsInternal._getLog(),
                        'version':env['version'],
                        'build':env['buildNumber'],
                        'userAgentString':bs.getEnvironment()['userAgentString'],
                        'session':sessionName,
                        'activity':activityName,
                        'fatal':0,
                        'userRanCommands':bsInternal._hasUserRunCommands(),
                        'time':bs.getRealTime(),
                        'userModded':bsInternal._hasUserMods()}

                def response(data):
                    # a non-None response means success; lets
                    # take note that we don't need to report further
                    # log info this run
                    if data is not None:
                        global gLogHaveNew
                        gLogHaveNew = False
                        bsInternal._markLogSent()
                bsUtils.serverPut('bsLog',info,response)
        gLogUploadTimerStarted = True

        # delay our log upload slightly in case other pertinant info gets printed
        # between now and then
        bs.realTimer(3000,_putLog)

        # after a while, allow another log-put
        def _reset():
            global gLogUploadTimerStarted
            gLogUploadTimerStarted = False
            if gLogHaveNew: _handleLog()
        if not bsInternal._isLogFull():
            bs.realTimer(1000*60*10,_reset)



def writeConfig(force=False,storeRemote=True):
    """category: General Utility Functions

    Commit the config to disk/cloud/etc"""
    
    global gConfigFileIsHealthy
    if not gConfigFileIsHealthy and not force:
        print "Current config file is broken; skipping write to avoid losing settings."
        return
    path = bs.getEnvironment()['configFilePath']
    pathPrev = path+'.prev'

    # write to a temp file then swap in.. this should give us a decently foolproof write
    # (though maybe we should sync before the rename?....)
    s = ''.join(_prettifyDict(bs.config))

    try:
        f=open(path+'.tmp','wb')
        f.write(s)
        f.close()
    except Exception,e:
        bs.screenMessage("Error writing config file:\n"+str(e))
        print ('Error writing config tmp file: '+str(e))
        return

    # temp - we're seeing corrupted configs on android, so lets
    # try to re-read the config immediately to get to the bottom of it
    try:
        fTest = open(path+'.tmp','rb')
        fTestRaw = fTest.read()
        fTest.close()
        fTestVals = eval(fTestRaw)
    except Exception,e:
        bs.screenMessage("Error writing config file (internal error)")
        print 'Error on config-write readback test:',e
        return

    # now move any existing config to .prev
    if os.path.exists(path):
        if 'windows' in bs.getEnvironment()['userAgentString']:
            try: os.remove(pathPrev)
            except Exception: pass
        try:
            os.rename(path,pathPrev)
        except Exception,e:
            bs.screenMessage("Error writing backup config file:\n"+str(e))
            print ('Error writing backup config file: '+str(e))
            return
            
    # ..and rename our temp file into place
    # on windows rename doesn't atomically overwrite existing files, so lets try to delete if one exists
    # if 'windows' in bs.getEnvironment()['userAgentString']:
    #     try: os.remove(path)
    #     except Exception: pass

    try:
        os.rename(path+'.tmp',path)
    except Exception,e:
        bs.screenMessage("Error renaming config file:\n"+str(e))
        print ('Error renaming config tmp file: '+str(e))
        return

    gConfigFileIsHealthy = True

    if storeRemote:
        # we want to send certain config keys to our remote config (if available)
        global gCloudConfig
        global gCloudConfigKeys
        configToSend = {}
        for key in gCloudConfigKeys:
            if key in bs.config:
                # we use our prettify func here instead of repr when possible
                # because it sorts dicts; otherwise our repr() may change even though the value hasn't
                if type(bs.config[key]) == dict:
                    valStr = ''.join(_prettifyDict(bs.config[key],0,False))
                elif type(bs.config[key]) == list:
                    valStr = ''.join(_prettifyList(bs.config[key],0,False))
                else: valStr = repr(bs.config[key])

                if key not in gCloudConfig or valStr != gCloudConfig[key]:
                    gCloudConfig[key] = configToSend[key] = valStr

        if len(configToSend) > 0:
            try:
                # we *always* send launchCount; we use it as a way to prevent overwriting
                # of new data with old
                configToSend['launchCount'] = repr(bs.config['launchCount'])
                bsInternal._storeRemoteConfig(configToSend)
            except Exception:
                print ('Error on remote config store: '+str(e))


def _setLanguage(language,printChange=True,storeToConfig=True):
    """Set the language used by the game.  Pass None to use OS default."""
    global _gLanguage

    try: curLanguage = bs.config['Lang']
    except Exception: curLanguage = None

    # store this in the config if its changing
    if language != curLanguage and storeToConfig:
        if language is None:
            del bs.config['Lang'] # clear it out for default
        else:
            bs.config['Lang'] = language
        writeConfig()
        switched = True
    else: switched = False

    # None implies default (currently locked to english)
    if language is None: language = _getDefaultLanguage()

    # create a single flattened dict for this language for simplicity
    languages = []

    try:
        l = __import__('bsLanguage'+language)
        if language != 'English': reload(l) # helpful for iterating
    except Exception,e:
        print 'Exception importing language module \'bsLanguage'+language+'\':',e
        bs.screenMessage("Error setting language to '"+language+"'; see log for details",color=(1,0,0))
        switched = False
        l = None

    if l is not None: languages.append(l)
    # everything inherits from english..
    if language != 'English': languages.append(__import__('bsLanguageEnglish'))

    languages.reverse()
    lFull = AttrDict()
    for l in languages:
        _addToAttrDict(lFull,l.values)
    _gLanguage = lFull

    # pass the root-level keys/values in for low level code to use
    bsInternal._setInternalLanguageKeys([v for v in lFull['internal'].items() if type(v[1]) in (str,unicode)])

    if switched and printChange:
        bs.screenMessage(bs.getResource('languageSetText').replace('${LANGUAGE}',bs.translate('languages',language)),color=(0,1,0))
        #bs.screenMessage("Language is now '"+language+"'.\n(Note: this will only affect new UI/Game elements)",color=(0,1,0))

class AttrDict(dict):
    """ a dict that can be accessed with dot notation: foo.bar is equivalant to foo['bar'] """
    def __getattr__(self, attr):
        return self[attr]
    def __setattr__(self, attr, value):
        self[attr] = value

def _addToAttrDict(dst,src):
    for key,value in src.items():
        if type(value) is dict:
            try: dstDict = dst[key]
            except Exception: dstDict = dst[key] = AttrDict()
            if type(dstDict) is not AttrDict: raise Exception("language key '"+key+"' is defined both as a dict and value")
            _addToAttrDict(dstDict,value)
        else:
            if type(value) not in (float,int,bool,str,unicode,type(None)): raise Exception("invalid value type for res '"+key+"': "+str(type(value)))
            dst[key] = value

def _getDefaultLanguage():
    l = bs.getEnvironment()['locale']
    if l.startswith('de'): return 'German'
    elif l.startswith('es'): return 'Spanish'
    elif l.startswith('it'): return 'Italian'
    elif l.startswith('nl'): return 'Dutch'
    elif l.startswith('da'): return 'Danish'
    elif l.startswith('pt'): return 'Portuguese'
    elif l.startswith('fr'): return 'French'
    elif l.startswith('ru'): return 'Russian'
    elif l.startswith('pl'): return 'Polish'
    return 'English'

def getLanguage(returnNoneForDefault=False):
    """
    category: General Utility Functions

    Returns the language currently being used by the game.
    This may or may not equal the language in use by the OS.
    """

    try: language = bs.config['Lang']
    except Exception: language = None

    if language is None and not returnNoneForDefault:
        language = _getDefaultLanguage()

    return language
    

        
def getResource(resource,default=None):
    """
    category: General Utility Functions

    Returns a named resource value.

    Resources are localized values; they are generally strings, but can be ints or other
    basic types. If a resource value is found for the current language, it will be returned.
    Otherwise, the English value will be returned.  If no English value is available,
    the value of the 'default' arg will be returned.  If no default arg was supplied,
    an exception will be raised.
    """
    try:
        global _gLanguage
        if _gLanguage is None:
            language = getLanguage()
            try: _setLanguage(language,printChange=False,storeToConfig=False)
            except Exception,e:
                print 'Error setting language to '+language+': ',e
                # try english as a fallback
                if (language != 'English'):
                    print 'Resorting to fallback language (English)'
                    try: _setLanguage('English',printChange=False,storeToConfig=False)
                    except Exception,e: print 'Error setting language to English fallback: ',e

        splits = resource.split('.')
        dicts = splits[:-1]
        key = splits[-1]
        values = _gLanguage
        for d in dicts: values = values[d]
        return values[key]

    except Exception,e:
        if default is not None: return default
        else: raise Exception("resource not found: '"+resource+"'")

def translate(category,s,raiseExceptions=False,printErrors=False):
    """
    category: General Utility Functions

    Given a category and string, returns a translated version of it.
    If a translation does not exist, the original string is returned.
    This is conceptually slightly different than bs.getResource() in
    that it operates on actual values; not predefined key names.
    This makes it more applicable to mods, since a database of common
    translations can be built up after-the-fact without requiring
    coordination between modders of what key-name corresponds with
    what phrase.
    """
    try: translated = bs.getResource('translations')[category][s]
    except Exception,e:
        if raiseExceptions: raise e
        if printErrors: print 'Translate error: category=\''+category+'\' name=\''+s+'\' exc='+str(e)+''
        translated = None
    if translated is None: translated = s
    return translated


def getTypeName(t):
    return t.__module__+'.'+t.__name__

def resolveTypeName(typeName):
    moduleName,className = typeName.split('.')
    module = __import__(moduleName)
    return getattr(module,className)
    
def getPlayerColors():
    """ return user-selectable player colors """
    return _gPlayerColors

# call suppressDebugReports() to keep from spamming the debug log server during development
_gSuppressDebugReports = False
def suppressDebugReports():
    bs.screenMessage("Supressing debug reports",color=(1,0,0))
    global _gSuppressDebugReports
    _gSuppressDebugReports = True
_gPrintedLiveObjectWarning = False

def animate(node,attr,keys,loop=False,offset=0,driver='gameTime'):
    """
    category: Game Flow Functions

    Creates an animCurve node with the provided values and time as an input,
    connect it to the provided attribute, and set it to die with the target.
    Key values are provided as time:value dictionary pairs.  Time values are relative
    to the current time. Returns the animCurve node.
    """

    if driver != 'gameTime': raise Exception("fixme; only support game-time currently")
    items = keys.items()
    items.sort()
    curve = bs.newNode("animCurve",owner=node,name='Driving '+str(node)+' \''+attr+'\'')
    bs.time.connectAttr(driver,curve,"in")
    curve.connectAttr("out",node,attr)
    curve.times = [time for time,val in items]
    curve.values = [val for time,val in items]
    curve.loop = loop
    curve.offset = bs.getGameTime()+offset
    # if we're not looping, set a timer to kill this curve after its done its job
    # FIXME - even if we are looping we should have a way to die once we get disconnected
    if loop == False:
        bs.gameTimer(int(items[-1][0])+1000,curve.delete)
    return curve

def animateArray(node,attr,size,keys,loop=False,offset=0,driver='gameTime'):
    """
    category: Game Flow Functions
    
    Like bsUtils.animate(), but operates on array attributes.
    """
    combine = bs.newNode('combine',owner=node,attrs={'size':size})
    if driver != 'gameTime': raise Exception("fixme; only support game-time currently")
    items = keys.items()
    items.sort()
    curves = {}
    for i in range(size):
        curve = bs.newNode("animCurve",owner=node,name='Driving '+str(node)+' \''+attr+'\' member '+str(i))
        bs.time.connectAttr(driver,curve,"in")
        curve.connectAttr("out",combine,'input'+str(i))
        curve.times = [time for time,val in items]
        curve.values = [val[i] for time,val in items]
        curve.loop = loop
        curve.offset = bs.getGameTime()+offset
        # if we're not looping, set a timer to kill this curve after its done its job
        if loop == False:
            bs.gameTimer(int(items[-1][0])+1000,curve.delete)
    combine.connectAttr('output',node,attr)
    # if we're not looping, set a timer to kill the combine once the job is done
    # FIXME - even if we are looping we should have a way to die once we get disconnected
    if loop == False:
        bs.gameTimer(int(items[-1][0])+1000,combine.delete)



# called internally..
def _shutdown():
    if _gMusicPlayer is not None:
        _gMusicPlayer.shutdown()
    # if _gITunesThread is not None:
    #     print 'CALLING SHUTDOWN'
    #     _gITunesThread.shutdown()



class MusicPlayer(object):

    def __init__(self):
        self._haveSetInitialVolume = False
        self._entryToPlay = None
        self._volume = 1.0
        self._actuallyPlaying = False

    def selectEntry(self,callback,currentEntry,selectionTargetName):
        return self.onSelectEntry(callback,currentEntry,selectionTargetName)
        
    def setVolume(self,volume):
        oldVolume = self._volume
        self._volume = volume
        self.onSetVolume(volume)
        self._updatePlayState()

    def _updatePlayState(self):
        # if we arent playing, should be, and have positive volume, do so
        if not self._actuallyPlaying:
            if self._entryToPlay is not None and self._volume > 0.0:
                self.onPlay(self._entryToPlay)
                self._actuallyPlaying = True
        else:
            if self._actuallyPlaying and (self._entryToPlay is None or self._volume <= 0.0):
                self.onStop()
                self._actuallyPlaying = False
                
    def play(self,entry):
        if not self._haveSetInitialVolume:
            self._volume = bsInternal._getSetting('Music Volume')
            self.onSetVolume(self._volume)
            self._haveSetInitialVolume = True
        self._entryToPlay = copy.deepcopy(entry)

        # if we're currently *actually* playing something, switch to the new thing..
        # otherwise update state which will start us playing *only* if proper (volume > 0, etc)
        if self._actuallyPlaying: self.onPlay(self._entryToPlay)
        else: self._updatePlayState()

    def stop(self):
        self._entryToPlay = None
        self._updatePlayState()

    def shutdown(self):
        self.onShutdown()
        
    # subclasses should override the following:
    def onSelectEntry(self,callback,currentEntry,selectionTargetName):
        'should present a GUI to select an entry; callback should be called with an entry or None for default'
        pass
    
    def onSetVolume(self,volume):
        'Called when the volume should be changed'
        pass

    def onPlay(self,entry):
        'Called when a new song/playlist/etc should be played'
        pass

    def onStop(self):
        'Called when the music should stop'
        pass

    def onShutdown(self):
        'Called on final app shutdown'
        pass



_gMusicPlayer = None
_gMusicPlayerType = None
#_gITunesThread = None

# for internal music player
def _getValidMusicFileExtensions():
    return ['mp3','ogg','m4a','wav','flac','mid']

if 'android' in bs.getEnvironment()['userAgentString']:

    # this talks to native stuff via our c layer
    class InternalMusicPlayer(MusicPlayer):
        def __init__(self):
            MusicPlayer.__init__(self)
            self._wantToPlay = False
            self._actuallyPlaying = False
            
        def onSelectEntry(self,callback,currentEntry,selectionTargetName):
            import bsUI
            return bsUI.GetSoundtrackEntryTypeWindow(callback,currentEntry,selectionTargetName)
            
        def onSetVolume(self,volume):
            bsInternal._musicPlayerSetVolume(volume);

        class _PickFolderSongThread(threading.Thread):
            def __init__(self,path,callback):
                threading.Thread.__init__(self)
                self._callback = callback
                self._path = path

            def run(self):
                try:
                    allFiles = []
                    validExtensions = ['.'+x for x in _getValidMusicFileExtensions()]
                    for root, subFolders, files in os.walk(self._path):
                        for f in files:
                            if any(f.lower().endswith(ext) for ext in validExtensions):
                                allFiles.insert(random.randrange(len(allFiles)+1),root+'/'+f)
                    if len(allFiles) == 0: raise Exception(bs.getResource('internal.noMusicFilesInFolderText'))
                    #bs.runInGameThread(bs.Call(self._callback,fileName=allFiles[random.randrange(len(allFiles))]))
                    bs.runInGameThread(bs.Call(self._callback,result=allFiles))
                except Exception,e:
                    bs.runInGameThread(bs.Call(self._callback,result=self._path,error=str(e)))
            
        def onPlay(self,entry):
            entryType = _getSoundtrackEntryType(entry)
            name = _getSoundtrackEntryName(entry)
            if entryType == 'musicFile':
                self._wantToPlay = self._actuallyPlaying = True
                bsInternal._musicPlayerPlay(name)
            elif entryType == 'musicFolder':
                # launch a thread to scan this folder and give us a random valid file within
                self._wantToPlay = True
                self._actuallyPlaying = False
                self._PickFolderSongThread(name,self._onPlayFolderCB).start()

        def _onPlayFolderCB(self,result,error=None):
            if error is not None:
                errStr = bs.getResource('internal.errorPlayingMusicText').replace('${MUSIC}',os.path.basename(result))+'; '+str(error)
                #print errStr
                bs.screenMessage(errStr,color=(1,0,0))
                return

            # theres a chance a stop could have been issued before our thread returned..
            # if that's the case, dont' play.
            if not self._wantToPlay:
                print '_onPlayFolderCB called with _wantToPlay False'
            else:
                self._actuallyPlaying = True
                bsInternal._musicPlayerPlay(result)
        
        def onStop(self):
            self._wantToPlay = False
            self._actuallyPlaying = False
            bsInternal._musicPlayerStop()

        def onShutdown(self):
            bsInternal._musicPlayerShutdown()

    _gMusicPlayerType = InternalMusicPlayer

elif 'Mac' in bs.getEnvironment()['userAgentString']:

    class ITunesThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self._commandsAvailable = threading.Event()
            self._commands = []
            self._volume = 1.0
            self._currentPlaylist = None

        def run(self):
            import ScriptingBridge as sb
            import LaunchServices as ls

            # iTunes seems to be locking up sometimes when bombsquad quits.
            # I'm guessing it might be related to the scripting bridge app going down
            # when this thread dies.
            # ..so as a workaround, lets make it a global so it won't get killed
            global gITunesApp
            try: gITunesApp
            except Exception:
                self._app = gITunesApp = sb.SBApplication.applicationWithBundleIdentifier_("com.apple.iTunes")
                # it looks like launching iTunes here on 10.7/10.8 knocks us out of fullscreen.
                # ew... that might be a bug, but for now we can work around it by reactivating
                # ourself after
                def doPrint():
                    bs.realTimer(1000,bs.Call(bs.screenMessage,bs.getResource('usingItunesText'),(0,1,0)))
                    bs.realTimer(3000,bs.Call(bs.screenMessage,bs.getResource('usingItunesTurnRepeatAndShuffleOnText'),(0,1,0)))

                bs.runInGameThread(doPrint)
                # here we grab this to force the actual launch
                # ..currently (on 10.8 at least) this is causing a switch away from our fullscreen window.
                # to work around this we explicitly focus our main window to bring ourself back
                self._app.soundVolume()
                bs.runInGameThread(bsInternal._focusWindow)

            self._source = self._app.sources().objectWithName_('Library')
            if not self._source.exists():
                self._source = self._app.sources()[0] # as a fallback just grab first
            done = False
            while not done:
                self._commandsAvailable.wait()
                self._commandsAvailable.clear()
                # (we're not protecting this list with a mutex but we're
                # just using it as a simple queue so it should be fine..)
                while len(self._commands) > 0:
                    cmd = self._commands.pop(0)
                    if cmd[0] == 'DIE':
                        # only stop if we've actually played something..
                        # (we don't want to be killing music the user has playing)
                        #if self._havePlayed:
                        if self._currentPlaylist and self._volume > 0:
                            try:
                                self._app.stop()
                                self._app.setSoundVolume_(self._origVolume)
                            except Exception,e: print 'Error stopping iTunes music:',e
                        done = True
                        break
                    elif cmd[0] == 'PLAY':
                        # None implies stop
                        if cmd[1] == None:
                            #if self._havePlayed:
                            if self._currentPlaylist and self._volume > 0:
                                try:
                                    self._app.stop()
                                    self._app.setSoundVolume_(self._origVolume)
                                except Exception,e: print 'Error stopping iTunes music:',e
                            self._currentPlaylist = None
                        else:
                            # if we've got something playing with positive volume, stop it
                            if self._currentPlaylist and self._volume > 0:
                                try:
                                    self._app.stop()
                                    self._app.setSoundVolume_(self._origVolume)
                                except Exception,e: print 'Error stopping iTunes music:',e
                            # set our playlist and play it if our volume is up
                            self._currentPlaylist = cmd[1]
                            if self._volume > 0:
                                self._origVolume = self._app.soundVolume()
                                self._updateITunesVolume()
                                self._playCurrentPlaylist()
                    elif cmd[0] == 'GET_PLAYLISTS':
                        try:
                            playlists = [p.name() for p in self._source.userPlaylists() if not p.name() in ['Music','Movies','TV Shows','Podcasts',u'iTunes\xa0U','Books','Genius','iTunes DJ', 'Music Videos']]
                            playlists.sort(key=lambda x: x.lower())
                        except Exception,e:
                            print 'Error getting iTunes playlists:',e
                            playlists = []
                        bs.runInGameThread(bs.Call(cmd[1],playlists))

                    del cmd # allows the command data/callback/etc to be freed

        def _playCurrentPlaylist(self):
            try:
                #print 'gonna do',self._currentPlaylist,type(self._currentPlaylist)
                playlistObj = self._source.userPlaylists().objectWithName_(self._currentPlaylist.decode('utf-8'))
                if playlistObj.exists():
                    # note: setting repeat and shuffle doesn't work in itunes 11 currently :-(
                    playlistObj.setSongRepeat_(1799449708) # repeat-all
                    playlistObj.setShuffle_(True)
                    playlistObj.playOnce_(False)
                else:
                    bs.runInGameThread(bs.Call(bs.screenMessage,bs.getResource('playlistNotFoundText')+': \''+self._currentPlaylist+'\'',(1,0,0)))
            except Exception,e:
                try: print 'Exception playing playlist "'+self._currentPlaylist+'":',e
                except Exception: pass

        def _updateITunesVolume(self):
            self._app.setSoundVolume_(max(0,min(100,int(100.0*self._volume))))

        def setVolume(self,volume):
            oldVolume = self._volume
            self._volume = volume

            # if we've got nothing we're supposed to be playing, don't touch itunes..
            if self._currentPlaylist == None: return

            # if volume is going to zero, stop actually playing but dont clear playlist
            if oldVolume > 0 and volume == 0:
                try:
                    self._app.stop()
                    self._app.setSoundVolume_(self._origVolume)
                except Exception,e: print 'Error stopping iTunes music:',e
            elif self._volume > 0:
                # if volume was zero, store pre-playing volume and start playing
                if oldVolume == 0: self._origVolume = self._app.soundVolume()
                self._updateITunesVolume()
                if oldVolume == 0: self._playCurrentPlaylist()

        def playPlaylist(self,musicType):
            self._commands.append(['PLAY',musicType])
            self._commandsAvailable.set()

        def shutdown(self):
            self._commands.append(['DIE'])
            self._commandsAvailable.set()
            self.join()

        def getPlaylists(self,callback):
            self._commands.append(['GET_PLAYLISTS',callback])
            self._commandsAvailable.set()


    # define a music-player that talks to iTunes and lets us
    # select playlists as entries
    class MacITunesMusicPlayer(MusicPlayer):
        def __init__(self):
            MusicPlayer.__init__(self)
            self._thread = ITunesThread()
            self._thread.start()

        def onSelectEntry(self,callback,currentEntry,selectionTargetName):
            import bsUI
            return bsUI.GetSoundtrackEntryTypeWindow(callback,currentEntry,selectionTargetName)

        def onSetVolume(self,volume):
            self._thread.setVolume(volume)

        def onPlay(self,entry):
            entryType = _getSoundtrackEntryType(entry)
            if entryType == 'iTunesPlaylist':
                self._thread.playPlaylist(_getSoundtrackEntryName(entry))
            else:
                print 'MacITunesMusicPlayer passed unrecognized entry type:',entryType

        def onStop(self):
            self._thread.playPlaylist(None)

        def onShutdown(self):
            self._thread.shutdown()

    _gMusicPlayerType = MacITunesMusicPlayer


def haveMusicPlayer():
    return (_gMusicPlayerType is not None)

def getMusicPlayer():
    global _gMusicPlayer
    if _gMusicPlayer is None:
        if _gMusicPlayerType is None: raise Exception("no music player type set")
        _gMusicPlayer = _gMusicPlayerType()
    return _gMusicPlayer

# def getITunesThread():
#     global _gITunesThread
#     if _gITunesThread == None:
#         volume = bsInternal._getSetting('Music Volume')
#         print 'CREATING WITH VOLUME',volume
#         _gITunesThread = ITunesThread(volume=volume)
#         _gITunesThread.start()
#     return _gITunesThread
    
def _musicVolumeChanged(val):
    if _gMusicPlayer is not None:
        _gMusicPlayer.setVolume(val)
        
    # if _gITunesThread is not None:
    #     print 'SETTING VOLUME',val
    #     _gITunesThread.setVolume(val)

gMusicMode = 'regular'
gMusicTypes = {'regular':None,'test':None}

def setMusicPlayMode(mode,forceRestart=False):
    global gMusicMode
    oldMode = gMusicMode
    gMusicMode = mode
    if oldMode != gMusicMode or forceRestart:
        # if we're switching into test mode we don't
        # actually play anything until its requested..
        # if we're switching *out* of test mode though
        # we want to go back to whatever the normal song was
        if mode == 'regular': playMusic(gMusicTypes['regular'])


def _isSoundtrackEntryTypeSupported(entryType):
    ua = bs.getEnvironment()['userAgentString']
    if entryType == 'iTunesPlaylist':
        return True if 'Mac' in ua else False
    elif entryType in ('musicFile','musicFolder'):
        return True if ('android' in ua and bsInternal._androidGetExternalStoragePath() is not None) else False
    elif entryType == 'default':
        return True
        
def _getSoundtrackEntryType(entry):
    'Given a soundtrack entry, returns its type, taking into account what is supported locally.'
    try:
        if entry is None: entryType = 'default'
        # simple string denotes iTunesPlaylist (legacy format)
        elif type(entry) in (str,unicode): entryType = 'iTunesPlaylist'
        # for other entries we expect type and name strings in a dict
        elif (type(entry) is dict
              and 'type' in entry and type(entry['type']) is str 
              and 'name' in entry and type(entry['name']) is str): entryType = entry['type']
        else: raise Exception("invalid soundtrack entry: "+str(entry)+" (type "+str(type(entry))+")")
        if _isSoundtrackEntryTypeSupported(entryType): return entryType
        else: raise Exception("invalid soundtrack entry:"+str(entry))
    except Exception,e:
        print 'EXC on _getSoundtrackEntryType',e
        return 'default'

def _getSoundtrackEntryName(entry):
    'Given a soundtrack entry, returns its name. Note that this does not ensure the entry is valid (use _getSoundtrackEntryType() for that).'
    try:
        if entry is None: return None
        # simple string denotes an iTunesPlaylist name (legacy entry format)
        elif type(entry) is unicode: return entry.encode('utf-8')
        elif type(entry) is str: return entry
        # for other entries we expect type and name strings in a dict
        elif (type(entry) is dict
              and 'type' in entry and type(entry['type']) is str 
              and 'name' in entry and type(entry['name']) is str): return entry['name']
        else: raise Exception("invalid soundtrack entry:"+str(entry))
    except Exception,e:
        print 'EXC on _getSoundtrackEntryName',e
        return 'default'

def playMusic(musicType):
    """
    category: Game Flow Functions

    A high level function to set or stop the current music based on a string musicType.
    
    Current valid values for 'musicType': 'Menu', 'Victory', 'CharSelect', 'RunAway', 'Onslaught',
    'Keep Away', 'Race', 'Epic Race', 'Scores', 'GrandRomp', 'ToTheDeath', 'Chosen One', 'ForwardMarch',
    'FlagCatcher', 'Survival', 'Epic', 'Sports', 'Hockey', 'Football', 'Flying', 'Scary', 'Marching'.

    This function will handle loading and playing sound media as necessary, and also supports custom
    user soundtracks on specific platforms so the user can override particular game music with their own.

    Pass None to stop music.
    """
    _playMusic(musicType)
    
def _playMusic(musicType,mode='regular',testSoundtrack=None):

    gMusicTypes[mode] = musicType
    
    # if the OS tells us there's currently music playing, all our operations default to playing nothing
    if bsInternal._isOSPlayingMusic(): musicType = None

    # if we're not in the mode this music is being set for, don't actually change what's playing
    if mode != gMusicMode: return

    # some platforms have a special music-player for things like iTunes soundtracks, mp3s, etc.
    # if this is the case, attempt to grab an entry for this music-type, and
    # if we have one, have the music-player play it.  If not, we'll play game music ourself
    if musicType is not None and _gMusicPlayerType is not None:
        try:
            if testSoundtrack is not None:soundtrack = testSoundtrack
            else: soundtrack = bs.config['Soundtracks'][bs.config['Soundtrack']]
            entry = soundtrack[musicType]
        except Exception: entry = None
    else: entry = None

    # go through music-player
    if entry is not None:

        # stop any existing internal music..
        global gMusic
        if gMusic is not None:
            gMusic.delete()
            gMusic = None

        # play music-player music..
        getMusicPlayer().play(entry)
        
    # handle via internal music...
    else:
        if musicType is not None:
            loop = True
            if musicType == 'Menu':
                filename = 'menuMusic'
                volume = 5.0
            elif musicType == 'Victory':
                filename = 'victoryMusic'
                volume = 6.0
                loop = False
            elif musicType == 'CharSelect':
                filename = 'charSelectMusic'
                volume = 2.0
            elif musicType == 'RunAway':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Onslaught':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Keep Away':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Race':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Epic Race':
                filename = 'slowEpicMusic'
                volume = 6.0
            elif musicType == 'Scores':
                filename = 'scoresEpicMusic'
                volume = 3.0
                loop = False
            elif musicType == 'GrandRomp':
                filename = 'grandRompMusic'
                volume = 6.0
            elif musicType == 'ToTheDeath':
                filename = 'toTheDeathMusic'
                volume = 6.0
            elif musicType == 'Chosen One':
                filename = 'survivalMusic'
                volume = 4.0
            elif musicType == 'ForwardMarch':
                filename = 'forwardMarchMusic'
                volume = 4.0
            elif musicType == 'FlagCatcher':
                filename = 'flagCatcherMusic'
                volume = 6.0
            elif musicType == 'Survival':
                filename = 'survivalMusic'
                volume = 4.0
            elif musicType == 'Epic':
                filename = 'slowEpicMusic'
                volume = 6.0
            elif musicType == 'Sports':
                filename = 'sportsMusic'
                volume = 4.0
            elif musicType == 'Hockey':
                filename = 'sportsMusic'
                volume = 4.0
            elif musicType == 'Football':
                filename = 'sportsMusic'
                volume = 4.0
            elif musicType == 'Flying':
                filename = 'flyingMusic'
                volume = 4.0
            elif musicType == 'Scary':
                filename = 'scaryMusic'
                volume = 4.0
            elif musicType == 'Marching':
                filename = 'whenJohnnyComesMarchingHomeMusic'
                volume = 4.0
            else:
                print "Unknown music: '"+musicType+"'"
                filename = 'flagCatcherMusic'
                volume = 6.0

        # stop any existing music-player playback
        if _gMusicPlayer is not None:
            _gMusicPlayer.stop()

        # stop any existing internal music..
        if gMusic is not None:
            gMusic.delete()
            gMusic = None

        # start up new internal music..
        if musicType is not None:
            # fixme - currently this won't start playing if we're paused
            # since attr values don't get updated until
            # node updates happen  :-(
            gMusic = bs.newNode(type='sound',
                                name='Global Music',
                                owner=bs,
                                attrs={'sound':bs.getSound(filename),
                                       'positional':False,
                                       'music':True,
                                       'volume':volume,
                                       'loop':loop})

class Spawner(object):
    """
    category: Game Flow Classes

    Creates a light flash and sends a bs.Spawner.SpawnMessage
    to the current activity after a delay.
    """
    
    class SpawnMessage(object):
        """
        category: Message Classes

        Spawn message sent by a bs.Spawner after its delay has passed.

        Attributes:

           spawner
              The bs.Spawner we came crom.

           data
              The data object passed by the user.

           pt
              The spawn position.
        """
        def __init__(self,spawner,data,pt):
            """
            Instantiate with the given values.
            """
            self.spawner = spawner
            self.data = data
            self.pt = pt

    def __init__(self,activity,data,pt,spawnTime):
        """
        Instantiate a spawner given some custom data,
        a position, and a spawn-time in milliseconds.
        """
        
        self._spawnerSound = bs.getSound('swip2')

        self._activity = weakref.ref(activity)
        self._data = data
        self._pt = pt

        # create a light where the spawn will happen
        self._light = bs.newNode('light',owner=activity,attrs={'translate':tuple(pt),'radius':0.1,'color':(1.0,0.1,0.1),'lightVolumes':False})

        s = float(spawnTime)/3750
        minVal = 0.4
        maxVal = 0.7
        bs.playSound(self._spawnerSound,position=(self._light,'translate'))
        animate(self._light,'intensity',{0:0.0,
                                         int(250*s):maxVal,      int(500*s):minVal,  int(750*s):maxVal,      int(1000*s):minVal,
                                         int(1250*s):1.1*maxVal, int(1500*s):minVal, int(1750*s):1.2*maxVal, int(2000*s):minVal,
                                         int(2250*s):1.3*maxVal, int(2500*s):minVal, int(2750*s):1.4*maxVal, int(3000*s):minVal,
                                         int(3250*s):1.5*maxVal, int(3500*s):minVal, int(3750*s):2.0,        int(4000*s):0.0})
        bs.gameTimer(int(spawnTime),self._spawn)

    def _spawn(self):

        bs.gameTimer(1000,self._light.delete)

        # only run if our activity still exists
        activity = self._activity()
        if activity: activity.handleMessage(self.SpawnMessage(self,self._data,self._pt))

def isPointInBox(p,b):
    """
    category: General Utility Functions

    Return whether a given point is within a given box.
    For use with standard def boxes (translate|rotate|scale).
    """
    return ((abs(p[0]-b[0]) <= b[6]*0.5) and (abs(p[1]-b[1]) <= b[7]*0.5) and (abs(p[2]-b[2]) <= b[8]*0.5))

def getPlayerProfileColors(profileName):
    """ given a profile, returns colors for them """

    try: color = bs.config['Player Profiles'][profileName]['color']
    except Exception:
        # key off name if possible
        if profileName is None: 
            color = _gPlayerColors[random.randrange(6)] # first 6 are bright-ish
        else:
            color = _gPlayerColors[sum([ord(c) for c in profileName]) % 6] # first 6 are bright-ish

    try: highlight = bs.config['Player Profiles'][profileName]['highlight']
    except Exception:
        # key off name if possible
        if profileName is None: 
            highlight = _gPlayerColors[random.randrange(len(_gPlayerColors)-2)] # last 2 are grey and white; ignore those or we get lots of old-looking people
        else:
            highlight = _gPlayerColors[sum([ord(c)+1 for c in profileName]) % (len(_gPlayerColors)-2)]

    return color,highlight

def getTimeString(t,centi=True):
    """
    category: General Utility Functions
    
    Given a value in milliseconds, returns a string with minutes:seconds:centiseconds.
    """
    m = str((t/1000)/60)
    s = str((t/1000)%60)
    cs = str(((t/10)%100))
    if len(s) == 1: s = '0'+s
    if len(cs) == 1: cs = '0'+cs
    if centi: return m+":"+s+"."+cs
    else: return m+":"+s

def showDamageCount(activity,damage,position,direction):
    lifespan = 1000
    t = bs.newNode('text',
                   owner=activity,
                   attrs={'text':damage,
                          'inWorld':True,
                          'hAlign':'center',
                          'color':(1,0.2,0.2,1),
                          'scale':[0.01]})
    # translate upward
    tcombine = bs.newNode("combine",owner=t)
    tcombine.connectAttr('output',t,'translate')
    vVals = []
    p = 0; v = 0.07
    count = 6
    for i in range(count):
        vVals.append((float(i)/count,p))
        p += v
        v *= 0.5
    pStart = position[0]; pDir = direction[0]
    animate(tcombine,"input0",dict([[i[0]*lifespan,pStart+pDir*i[1]] for i in vVals]))

    pStart = position[1]; pDir = direction[1]
    animate(tcombine,"input1",dict([[i[0]*lifespan,pStart+pDir*i[1]] for i in vVals]))

    pStart = position[2]; pDir = direction[2]
    animate(tcombine,"input2",dict([[i[0]*lifespan,pStart+pDir*i[1]] for i in vVals]))
    ccombine = bs.newNode("combine",owner=t,attrs={'input0':1.0,'input1':0.4,'input2':0.4})

    ccombine.connectAttr('output',t,'color')
    animate(ccombine,"input3",{0.7*lifespan:1.0,lifespan:0.0})
    bs.gameTimer(lifespan,t.delete)

def getLastPlayerNameFromInputDevice(device):
    'Returns a reasonable player name associated with a device (generally the last one used there)'
    # look for a default player profile name for them.. otherwise default to their current random name
    profileName = '_random'
    keyName = device.getName()+' '+device.getUniqueIdentifier()
    if 'Default Player Profiles' in bs.config and keyName in bs.config['Default Player Profiles']:
        profileName = bs.config['Default Player Profiles'][keyName]
    if profileName == '_random': profileName = device._getDefaultPlayerName()
    return profileName

def getNormalizedColor(color):
    """
    category: General Utility Functions

    Scale a color so its largest value is 1; useful for coloring lights.
    """
    colorBiased = tuple(max(c,0.01) for c in color) # account for black and whatnot
    mult = 1.0/max(colorBiased)
    return tuple(c*mult for c in colorBiased)


# def getSafeDisplayColor(color):
#     """ given a color, returns one that is safe to display over black """

#     c = bs.getSafeColor(color=(color[0],color[1],color[2]));
#     if len(color) == 4: return c+(float(color[3]),)
#     else: return c


def garbageCollect():
    gc.collect()
    # can be handy to print this to check for leaks between games
    if 0: print 'PY OBJ COUNT',len(gc.get_objects())
    if len(gc.garbage) > 0:
        print 'PTHON GC FOUND',len(gc.garbage),'UNCOLLECTABLE OBJECTS:'
        for i,obj in enumerate(gc.garbage):
            print str(i)+':',obj

    printLiveObjectWarnings('after session shutdown')

def printLiveObjectWarnings(when,ignoreSession=None,ignoreActivity=None):
    sessions = []
    activities = []
    actors = []
    materials = []
    global _gPrintedLiveObjectWarning
    
    if _gPrintedLiveObjectWarning:
        # print 'skipping live object check due to previous found live object(s)'
        return

    for obj in gc.get_objects():
        if isinstance(obj,bs.Actor): actors.append(obj)
        elif isinstance(obj,bs.Session): sessions.append(obj)
        elif isinstance(obj,bs.Activity): activities.append(obj)
        elif isinstance(obj,bs.Material): materials.append(obj)

    # complain about any remaining sessions
    for session in sessions:
        if session is ignoreSession: continue
        _gPrintedLiveObjectWarning = True
        print 'Session found',when,':',type(session)
        refs = list(gc.get_referrers(session))
        i = 1
        for ref in refs:
            if type(ref) is types.FrameType: continue
            print '     ref',i,':',ref
            i += 1
            if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                refs2 = list(gc.get_referrers(ref))
                j = 1
                for ref2 in refs2:
                    if type(ref2) is types.FrameType: continue
                    print '        ref\'s ref',j,':',ref2
                    j += 1

    # complain about any remaining activities
    for activity in activities:
        if activity is ignoreActivity: continue
        _gPrintedLiveObjectWarning = True
        print 'Activity found',when,':',type(activity)
        refs = list(gc.get_referrers(activity))
        i = 1
        for ref in refs:
            if type(ref) is types.FrameType: continue
            print '     ref',i,':',ref
            i += 1
            if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                refs2 = list(gc.get_referrers(ref))
                j = 1
                for ref2 in refs2:
                    if type(ref2) is types.FrameType: continue
                    print '        ref\'s ref',j,':',ref2
                    j += 1

    # complain about any remaining actors
    for actor in actors:
        _gPrintedLiveObjectWarning = True
        print 'Actor found',when,':',type(actor)
        if type(actor) is bs.Actor:
            try:
                if actor.node.exists(): print '   - contains node:',actor.node.getNodeType(),';',actor.node.getName()
            except Exception,e:
                print '   - exception checking actor node:',e
        refs = list(gc.get_referrers(actor))
        i = 1
        for ref in refs:
            if type(ref) is types.FrameType: continue
            print '     ref',i,':',ref
            i += 1
            if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                refs2 = list(gc.get_referrers(ref))
                j = 1
                for ref2 in refs2:
                    if type(ref2) is types.FrameType: continue
                    print '        ref\'s ref',j,':',ref2
                    j += 1

    # complain about any remaining materials
    for material in materials:
        _gPrintedLiveObjectWarning = True
        print 'Material found',when,':',material.getName()
                    
    # complain about any remaining nodes
    j = 1
    for node in bs.getNodes():
        if _isValidRemainingNode(node): continue
        _gPrintedLiveObjectWarning = True
        print 'Live node found',when,':',j,':',node.getNodeType(),';',node.getName()
        j += 1

def _isValidRemainingNode(node):
    return (node.getName() in ['Globals','Global Music','Global Time'])

def _weakDeleteNode(node,weakRef):
    node.delete()

def _setNodeOwner(node,owner,isCrossFade):

    # ..owner can be an activity..
    if isinstance(owner,bs.Activity):
        owner._addOwnedNode(node,isCrossFade)

    # a few key nodes need to never get auto-deleted; they pass bs in this case
    # (hmm i suppose the weak-ref case would cover this too..)
    elif owner is bs:
        pass

    # for all other types, create a weak-ref callback to kill the node when the target object dies
    # hmm should we explicitly limit this to actors or something?
    else:
        global gNodeOwnerWeakRefs
        global gNodeOwnerWeakRefsCleanCounter
        gNodeOwnerWeakRefs.append(weakref.ref(owner,bs.Call(_weakDeleteNode,node)))
        gNodeOwnerWeakRefsCleanCounter += 1
        # pull dead refs off our list periodically (I supposed the callback could do it, but this works)
        if (gNodeOwnerWeakRefsCleanCounter > 10):
            gNodeOwnerWeakRefs = [r for r in gNodeOwnerWeakRefs if r() is not None]
            gNodeOwnerWeakRefsCleanCounter = 0
    
class ServerCallThread(threading.Thread):
    def __init__(self,request,requestType,data,callback):
        threading.Thread.__init__(self)
        self._request = request
        self._requestType = requestType
        self._data = {} if data == None else data
        self._callback = callback

    def run(self):
        try:
            if self._requestType == 'get':
                response = urllib2.urlopen(urllib2.Request(bsInternal._getServerAddress()+'/'+self._request+'?'+urllib.urlencode(self._data),
                                                           None,
                                                           { 'User-Agent' : bs.getEnvironment()['userAgentString'] }))
            elif self._requestType == 'post':
                response = urllib2.urlopen(urllib2.Request(bsInternal._getServerAddress()+'/'+self._request,
                                                           urllib.urlencode(self._data),
                                                           { 'User-Agent' : bs.getEnvironment()['userAgentString'] }))
            else: raise Exception("Invalid requestType: "+self._requestType)
            responseData = ast.literal_eval(response.read())
            if self._callback is not None: bs.runInGameThread(bs.Call(self._callback,responseData))
        except Exception,e:
            if self._callback is not None: bs.runInGameThread(bs.Call(self._callback,None))


def serverGet(request,data,callback=None):
    ServerCallThread(request,'get',data,callback).start()

def serverPut(request,data,callback=None):
    ServerCallThread(request,'post',data,callback).start()

def runStressTest(playlistType='Random',playlistName='__default__',playerCount=8,roundDuration=30):
    bs.screenMessage('Beginning stress test.. use \'End Game\' to stop testing.',color=(1,1,0))
    bs.realTimer(1000,bs.Call(_startStressTest,{'playlistType':playlistType,
                                                'playlistName':playlistName,
                                                'playerCount':playerCount,
                                                'roundDuration':roundDuration}))
    bs.realTimer(8000,bs.Call(bs.screenMessage,'stats will be written to '+getHumanReadableUserScriptsPath()+'/stressTestStats.csv'))

def stopStressTest():
    bsInternal._setStressTesting(False,0)
    global _gStressTestResetTimer
    try:
        if _gStressTestResetTimer is not None:
            bs.screenMessage("Ending stress test...",color=(1,1,0))
    except Exception: pass
    _gStressTestResetTimer = None

def _startStressTest(args):
    playlistType = args['playlistType']
    if playlistType == 'Random':
        if (random.random() < 0.5): playlistType = 'Teams'
        else: playlistType = 'Free-For-All'
    bs.screenMessage('Running Stress Test (listType="'+playlistType+'", listName="'+args['playlistName']+'")...')
    if playlistType == 'Teams':
        bs.config['Team Tournament Playlist Selection'] = args['playlistName']
        bs.config['Team Tournament Playlist Randomize'] = 1
        bsInternal._hostSession('bsTeamGame.TeamsSession')
    else:
        bs.config['Free-for-All Playlist Selection'] = args['playlistName']
        bs.config['Free-for-All Playlist Randomize'] = 1
        bsInternal._hostSession('bsTeamGame.FreeForAllSession')
    bsInternal._setStressTesting(True,args['playerCount'])
    global _gStressTestResetTimer
    _gStressTestResetTimer = bs.Timer(args['roundDuration']*1000,bs.Call(_resetStressTest,args),timeType='real')


def _resetStressTest(args):
    bsInternal._setStressTesting(False,args['playerCount'])
    bs.screenMessage('Resetting stress test...')
    bs.getSession().end()
    bs.realTimer(2000,bs.Call(_startStressTest,args))

class ControlsHelpOverlay(bsGame.Actor):
    """
    category: Game Flow Classes

    A screen overlay of game controls, showing button mappings
    based on what controllers are connected. Handy to show at the
    start of a series or whenever there might be newbies watching.
    """
    def __init__(self,activity,retain,position=(390,120),scale=1.0,delay=0,lifespan=None,bright=False):
        """
        Instantiate an overlay.

        delay: is the time in milliseconds before the overlay fades in.

        lifespan: if not None, the overlay will fade back out and die after that long (in milliseconds)

        bright: if True, brighter colors will be used; handy when showing over gameplay but may be
                too bright for join-screens, etc.
        """
        bsGame.Actor.__init__(self,activity,retain)

        showTitle = True
        
        scale *= 0.75
        
        imageSize = 90.0*scale
        
        offs = 74.0*scale
        offs5 = 43.0*scale
        ouya = bsInternal._isOuyaVersion()
        mw = 50

        shadOffs = 2.0*scale

        self._lifespan = lifespan
        self._dead = False
        
        if showTitle:
            self._titleTextPosTop = (position[0],position[1]+139.0*scale)
            self._titleTextPosBottom = (position[0],position[1]+139.0*scale)
            c = (1,1,1) if bright else (0.7,0.7,0.7)
            self._titleText = bs.newNode('text',owner=activity,
                                         attrs={'text':bs.getResource('controlsText')+':',
                                                'scale':[1.1*scale],
                                                'shadow':1.0,
                                                'maxWidth':480,
                                                'vAlign':'center','hAlign':'center',
                                                'color':c})
        else: self._titleText = None

        p = (position[0],position[1]-offs)
        c = (0.4,1,0.4)
        self._jumpImage = bs.newNode('image',owner=activity,
                               attrs={'texture':bs.getTexture('buttonJump'),'absoluteScale':True,
                                       'translate':p,'scale':(imageSize,imageSize),'color':c})
        self._jumpText = bs.newNode('text',owner=activity,
                              attrs={'vAlign':'top','hAlign':'center','scale':[1.5*scale],
                                     'shadow':1.0, 'maxWidth':mw,'translate':(p[0],p[1]-offs5),'color':c})


        p = (position[0]-offs*1.1,position[1])
        c = (0.2,0.6,1) if ouya else (1,0.7,0.3)
        self._punchImage = bs.newNode('image',owner=activity,
                               attrs={'texture':bs.getTexture('buttonPunch'),'absoluteScale':True,
                                      'translate':p,'scale':(imageSize,imageSize),'color':c})
        self._punchText = bs.newNode('text',owner=activity,
                               attrs={'vAlign':'top','hAlign':'center','scale':[1.5*scale],
                                      'shadow':1.0, 'maxWidth':mw,'translate':(p[0],p[1]-offs5),'color':c})

        p = (position[0]+offs*1.1,position[1])
        c = (1,0.3,0.3)
        self._bombImage = bs.newNode('image',owner=activity,
                               attrs={'texture':bs.getTexture('buttonBomb'),'absoluteScale':True,
                                      'translate':p,'scale':(imageSize,imageSize),'color':c})
        self._bombText = bs.newNode('text',owner=activity,
                               attrs={'hAlign':'center','vAlign':'top','scale':[1.5*scale],
                                      'shadow':1.0, 'maxWidth':mw,'translate':(p[0],p[1]-offs5),'color':c})

        p = (position[0],position[1]+offs)
        c = (1,0.8,0.3) if ouya else (0.8,0.5,1)
        self._pickUpImage = bs.newNode('image',owner=activity,
                                 attrs={'texture':bs.getTexture('buttonPickUp'),'absoluteScale':True,
                                        'translate':p,'scale':(imageSize,imageSize),'color':c})
        self._pickUpText = bs.newNode('text',owner=activity,
                                attrs={'vAlign':'top','hAlign':'center','scale':[1.5*scale],
                                       'shadow':1.0, 'maxWidth':mw,'translate':(p[0],p[1]-offs5),'color':c})
        
        
        c = (0.9,0.9,2.0) if bright else (0.8,0.8,2.0,1.0)
        self._runTextPosTop = (position[0],position[1] - 135.0*scale)
        self._runTextPosBottom = (position[0],position[1] - 172.0*scale)
        self._runText = bs.newNode('text',owner=activity,
                                attrs={'scale':[0.8*scale],
                                       'shadow':1.0,
                                       'maxWidth':480,
                                       'vAlign':'top','hAlign':'center',
                                       'color':c})

        self._nodes = [self._bombImage,self._bombText,
                       self._punchImage,self._punchText,
                       self._jumpImage,self._jumpText,
                       self._pickUpImage,self._pickUpText,
                       self._runText]
        if showTitle: self._nodes.append(self._titleText)

        # start everything invisible
        for n in self._nodes:
            n.opacity = 0.0

        # dont do anything until our delay has passed
        bs.gameTimer(delay,bs.WeakCall(self._startUpdating))

    def _startUpdating(self):

        # ok, our delay has passed.. now lets periodically see if we can fade in
        # (if a touch-screen is present we only want to show up if gamepads are connected, etc)
        
        # also set up a timer so if we havnt faded in by the end of our duration, abort.
        if self._lifespan is not None:
            self._cancelTimer = bs.Timer(self._lifespan,bs.WeakCall(self.handleMessage,bs.DieMessage(immediate=True)))
        
        self._fadeInTimer = bs.Timer(1000,bs.WeakCall(self._checkFadeIn),repeat=True)
        self._checkFadeIn() # do one check immediately

    def _checkFadeIn(self):
        import bsUI
        
        # if we have a touchscreen, we only fade in if we have a player with 
        # an input device that is *not* the touchscreen
        try: touchscreen = bsInternal._getInputDevice('TouchScreen','#1')
        except Exception: touchscreen = None
        if touchscreen is not None:
            # we look at the session's players; not the activity's - we want to get ones
            # who are still in the process of selecting a character, etc.
            inputDevices = [p.getInputDevice() for p in self.getActivity().getSession().players]
            inputDevices = [i for i in inputDevices if i is not None and i.exists() and i is not touchscreen]
            fadeIn = False
            if len(inputDevices) > 0:
                # only count this one if it has non-empty button names
                # (filters out wiimotes, the remote-app, etc)
                for device in inputDevices:
                    for name in ('buttonPunch','buttonJump','buttonBomb','buttonPickUp'):
                        if device.getButtonName(bsUI.getControllerValue(device,name)) != '':
                            fadeIn = True
                            break
                    if fadeIn: break # no need to keep looking

        else:
            # no touch-screen; fade in immediately
            fadeIn = True

        if fadeIn:
            self._cancelTimer = None # didnt need this
            self._fadeInTimer = None # done with this
            self._fadeIn()

        
    def _fadeIn(self):
        for n in self._nodes:
            bs.animate(n,'opacity',{0:0.0,2000:1.0})

        # if we were given a lifespan, transition out after it..
        if self._lifespan is not None:
            bs.gameTimer(self._lifespan,bs.WeakCall(self.handleMessage,bs.DieMessage()))

        self._update()
        self._updateTimer = bs.Timer(1000,bs.WeakCall(self._update),repeat=True,timeType='real')
        
    def _update(self):
        import bsUI

        punchButtonNames = set()
        jumpButtonNames = set()
        pickUpButtonNames = set()
        bombButtonNames = set()

        # we look at the session's players; not the activity's - we want to get ones
        # who are still in the process of selecting a character, etc.
        inputDevices = [p.getInputDevice() for p in self.getActivity().getSession().players]
        inputDevices = [i for i in inputDevices if i is not None and i.exists()]

        # if there's no players with input devices yet, try to default to showing keyboard controls
        if len(inputDevices) == 0:
            try: kb = bsInternal._getInputDevice('Keyboard','#1')
            except Exception: kb = None
            if kb is not None: inputDevices.append(kb)
        
        # we word things specially if we have nothing but keyboards..
        allKeyboards = (len(inputDevices) > 0
                        and all(i.getName() == 'Keyboard' for i in inputDevices))

        if allKeyboards:
            rightButtonNames = set()
            leftButtonNames = set()
            upButtonNames = set()
            downButtonNames = set()
            
        # for each player in the game with an input device,
        # get the name of the button for each of these 4 actions.
        # if any of them are uniform across all devices, display the name
        for device in inputDevices:
            # (we only care about movement buttons in the case of keyboards)
            if allKeyboards:
                rightButtonNames.add(device.getButtonName(bsUI.getControllerValue(device,'buttonRight')))
                leftButtonNames.add(device.getButtonName(bsUI.getControllerValue(device,'buttonLeft')))
                downButtonNames.add(device.getButtonName(bsUI.getControllerValue(device,'buttonDown')))
                upButtonNames.add(device.getButtonName(bsUI.getControllerValue(device,'buttonUp')))
            # ignore empty values; things like the remote app or wiimotes can return these..
            bname = device.getButtonName(bsUI.getControllerValue(device,'buttonPunch'))
            if bname != '': punchButtonNames.add(bname)
            bname = device.getButtonName(bsUI.getControllerValue(device,'buttonJump'))
            if bname != '': jumpButtonNames.add(bname)
            bname = device.getButtonName(bsUI.getControllerValue(device,'buttonBomb'))
            if bname != '': bombButtonNames.add(bname)
            bname = device.getButtonName(bsUI.getControllerValue(device,'buttonPickUp'))
            if bname != '': pickUpButtonNames.add(bname)

        # if we have no values yet, we may want to throw out some sane defaults...
        if all(len(l) == 0 for l in (punchButtonNames,jumpButtonNames,bombButtonNames,pickUpButtonNames)):
            # show ouya buttons
            if bsInternal._isOuyaVersion():
                # FIXME - should look at the ouya controller config; not just list defaults
                punchButtonNames.add(bs.getSpecialChar('ouyaButtonU'))
                jumpButtonNames.add(bs.getSpecialChar('ouyaButtonO'))
                bombButtonNames.add(bs.getSpecialChar('ouyaButtonA'))
                pickUpButtonNames.add(bs.getSpecialChar('ouyaButtonY'))
            # otherwise on android show standard buttons
            elif bs.getEnvironment()['platform'] == 'android':
                punchButtonNames.add('X')
                jumpButtonNames.add('A')
                bombButtonNames.add('B')
                pickUpButtonNames.add('Y')
            
        runText = (bs.getResource('runText')+': '
                   +bs.getResource('holdAnyKeyText' if allKeyboards else 'holdAnyButtonText'))
        # if we're all keyboards, lets show move keys too
        if (allKeyboards and len(upButtonNames) == 1 and len(downButtonNames) == 1
            and len(leftButtonNames) == 1 and len(rightButtonNames) == 1):
            upText = list(upButtonNames)[0]
            downText = list(downButtonNames)[0]
            leftText = list(leftButtonNames)[0]
            rightText = list(rightButtonNames)[0]
            runText = bs.getResource('moveText')+': '+upText+', '+leftText+', '+downText+', '+rightText+'\n'+runText
            
        self._runText.text = (runText)


        if len(punchButtonNames) == 1: self._punchText.text = list(punchButtonNames)[0]
        else: self._punchText.text = ''

        if len(jumpButtonNames) == 1: t = list(jumpButtonNames)[0]
        else: t = ''
        self._jumpText.text = t
        
        if t == '': self._runText.translate = self._runTextPosTop
        else: self._runText.translate = self._runTextPosBottom
            
        if len(bombButtonNames) == 1: self._bombText.text = list(bombButtonNames)[0]
        else: self._bombText.text = ''

        # also move our title up/down depending on if this is shown.
        if len(pickUpButtonNames) == 1:
            self._pickUpText.text = list(pickUpButtonNames)[0]
            if self._titleText is not None:
                self._titleText.translate = self._titleTextPosTop
        else:
            self._pickUpText.text = ''
            if self._titleText is not None: self._titleText.translate = self._titleTextPosBottom
        

    def _die(self):
        for node in self._nodes: node.delete()
        self._nodes = []
        self._updateTimer = None
        self._dead = True

    def exists(self):
        return not self._dead
    
    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage):
            if m.immediate:
                self._die()
            else:
                # if they dont need immediate, fade out our nodes and die later
                for node in self._nodes:
                    bs.animate(node,'opacity',{0:node.opacity,3000:0.0})
                bs.gameTimer(3100,bs.WeakCall(self._die))
                
        else: bsGame.Actor.handleMessage(self,m)

class Background(bsGame.Actor):
    'Simple Fading Background Actor'

    def __init__(self,activity,retain,fadeTime=500,startFaded=False,showLogo=False):
        bs.Actor.__init__(self,activity,retain)
        self._dying = False
        self.fadeTime=fadeTime
        self.node = bs.newNode('image',
                               owner=activity,
                               delegate=self,
                               attrs={'fillScreen':True,
                                      'texture':bs.getTexture('bg'),
                                      'tiltTranslate':-0.3,
                                      'hasAlphaChannel':False,
                                      'color':(1,1,1)},
                               name='Background',
                               isCrossFade=True)

        if not startFaded:
            animate(self.node,'opacity',{0:0, self.fadeTime:1},loop=False)

        if showLogo:
            logoTexture = bs.getTexture('logo')
            logoModel = bs.getModel('logo')
            logoModelTransparent = bs.getModel('logoTransparent')
            self.c = bs.newNode('combine',owner=self.node,attrs={'size':2},name='Logo Scale Driver')
            for attr in ['input0','input1']:
                animate(self.c,attr,{0:0.693, 50:0.7, 500:0.693},loop=True)
            self.logo = bs.newNode('image',
                                   owner=self.node,
                                   attrs={'texture':logoTexture,
                                          'modelOpaque':logoModel,
                                          'modelTransparent':logoModelTransparent,
                                          'scale':(1,0.7),
                                          'color':(0.15,0.15,0.15),
                                          'translate':(0,0),
                                          'tiltTranslate':-0.05,
                                          'absoluteScale':False},name='Logo')
            self.c.connectAttr('output',self.logo,'scale')
            self.node.connectAttr('opacity',self.logo,'opacity')

            c = bs.newNode('combine',owner=self.node,name='Logo Translate Driver')
            c.connectAttr('output',self.logo,'translate')

            # gen some random keys for that stop-motion-y look
            keys = {}
            time = 0
            for i in range(10):
                keys[time] = (random.random()-0.5)*0.0015
                time += random.random() * 100
            animate(c,"input0",keys,loop=True)
            keys = {}
            time = 0
            for i in range(10):
                keys[time] = (random.random()-0.5)*0.0015 + 0.05
                time += random.random() * 100
            animate(c,"input1",keys,loop=True)


    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage):
            if not self._dying and self.node.exists():
                self._dying = True
                animate(self.node,"opacity",{0:1, self.fadeTime:0},loop=False)
                bs.gameTimer(self.fadeTime+100,self.node.delete)
        else: bs.Actor.handleMessage(self,m)


class Text(bsGame.Actor):
    """ Text with some tricks """

    def __init__(self,activity,retain,text,position=(0,0),hAlign='left',vAlign='none',color=(1,1,1,1),transition=None,transitionDelay=0,flash=False,
                 vAttach='center',hAttach='center',scale=1.0,transitionOutDelay=None,maxWidth=None,shadow=0.5):
        bs.Actor.__init__(self,activity,retain)
        self.node = bs.newNode('text',
                               name="Announce Text: '"+text+"'",
                               owner=activity,
                               delegate=self,
                               attrs={'text':text,
                                      'color':color,
                                      'translate':position,
                                      'hAlign':hAlign,
                                      'vAlign':vAlign,
                                      'hAttach':hAttach,
                                      'vAttach':vAttach,
                                      'shadow':shadow,
                                      'maxWidth':0.0 if maxWidth is None else maxWidth,
                                      'scale':[scale]})

        if transition == 'fadeIn':
            if flash: raise Exception("fixme: flash and fade-in currently cant both be on")
            c = bs.newNode('combine',owner=self.node,attrs={'input0':color[0],'input1':color[1],'input2':color[2]})
            c.connectAttr('output',self.node,'color')
            keys = {transitionDelay:0,transitionDelay+500:color[3]}
            if transitionOutDelay is not None:
                keys[transitionDelay+transitionOutDelay] = color[3]
                keys[transitionDelay+transitionOutDelay+500] = 0
            animate(c,"input3",keys)

        if flash:
            c = bs.newNode('combine',owner=self.node)
            c.connectAttr('output',self.node,'color')
            animate(c,"input0",{0:color[0], 150:0.3},loop=True)
            animate(c,"input1",{0:color[1], 150:0.1},loop=True)
            c.input2 = color[2]
            c.input3 = color[3]

        c = self.translateCombine = bs.newNode('combine',owner=self.node)
        c.connectAttr('output',self.node,'translate')
        if transition == 'inRight':
            keys = {transitionDelay:position[0]+1300,transitionDelay+200:position[0]}
            animate(c,'input0',keys)
            c.input1 = position[1]
        elif transition == 'inLeft':
            keys = {transitionDelay:position[0]-1300,transitionDelay+200:position[0]}
            if transitionOutDelay is not None:
                keys[transitionDelay+transitionOutDelay] = position[0]
                keys[transitionDelay+transitionOutDelay+200] = position[0]-1300
            animate(c,'input0',keys)
            c.input1 = position[1]
        elif transition == 'inBottomSlow':
            keys = {transitionDelay:-50,transitionDelay+1000:position[1]}
            c.input0 = position[0]
            animate(c,'input1',keys)
        elif transition == 'inBottom':
            keys = {transitionDelay:-50,transitionDelay+200:position[1]}
            if transitionOutDelay is not None:
                keys[transitionDelay+transitionOutDelay] = position[1]
                keys[transitionDelay+transitionOutDelay+200] = -50
            c.input0 = position[0]
            animate(c,'input1',keys)
        elif transition == 'inTopSlow':
            keys = {transitionDelay:400,transitionDelay+3500:position[1]}
            c.input0 = position[0]
            animate(c,'input1',keys)
        else:
            c.input0 = position[0]
            c.input1 = position[1]

        # if we're transitioning out, die at the end of it
        if transitionOutDelay is not None:
            bs.gameTimer(transitionDelay+transitionOutDelay+1000,bs.WeakCall(self.handleMessage,bs.DieMessage()))

    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage):
            self.node.delete()
        else:
            bs.Actor.handleMessage(self,m)


class Image(bsGame.Actor):

    """ just an image node with a few tricks """
    def __init__(self,activity,retain,texture,position=(0,0),transition=None,transitionDelay=0,attach='center',
                 color=(1,1,1,1),scale=(100,100),transitionOutDelay=None,modelOpaque=None,modelTransparent=None):
        bs.Actor.__init__(self,activity,retain)

        # if they provided a dict as texture, assume its an icon.. otherwise its just a texture value itself
        if type(texture) == dict:
            tintColor = texture['tintColor']
            tint2Color = texture['tint2Color']
            tintTexture = texture['tintTexture']
            texture = texture['texture']
        else:
            tintColor=(1,1,1)
            tint2Color=None
            tintTexture = -1
        

        self.node = bs.newNode('image',
                               owner=activity,
                               attrs={'texture':texture,
                                      'tintColor':tintColor,
                                      'tintTexture':tintTexture,
                                      'maskTexture':bs.getTexture('characterIconMask'),
                                      'translate':position,
                                      'scale':scale,
                                      'color':color[:3],
                                      'opacity':color[3],
                                      'absoluteScale':True,
                                      'attach':attach},
                               delegate=self)

        if modelOpaque is not None: self.node.modelOpaque = modelOpaque
        if modelTransparent is not None: self.node.modelTransparent = modelTransparent
        if tint2Color is not None: self.node.tint2Color = tint2Color

        if transition == 'fadeIn':
            keys = {transitionDelay:0,transitionDelay+500:color[3]}
            if transitionOutDelay is not None:
                keys[transitionDelay+transitionOutDelay] = color[3]
                keys[transitionDelay+transitionOutDelay+500] = 0
            animate(self.node,'opacity',keys)

        c = self.translateCombine = bs.newNode('combine',owner=self.node)
        c.connectAttr('output',self.node,'translate')
        if transition == 'inRight':
            keys = {transitionDelay:position[0]+1200,transitionDelay+200:position[0]}
            animate(c,'input0',keys)
            c.input1 = position[1]
        elif transition == 'inLeft':
            keys = {transitionDelay:position[0]-1200,transitionDelay+200:position[0]}
            if transitionOutDelay is not None:
                keys[transitionDelay+transitionOutDelay] = position[0]
                keys[transitionDelay+transitionOutDelay+200] = -position[0]-1200
            animate(c,'input0',keys)
            c.input1 = position[1]
        elif transition == 'inBottomSlow':
            keys = {transitionDelay:-400,transitionDelay+3500:position[1]}
            c.input0 = position[0]
            animate(c,'input1',keys)
        elif transition == 'inBottom':
            keys = {transitionDelay:-400,transitionDelay+200:position[1]}
            if transitionOutDelay is not None:
                keys[transitionDelay+transitionOutDelay] = position[1]
                keys[transitionDelay+transitionOutDelay+200] = -400
            c.input0 = position[0]
            animate(c,'input1',keys)
        elif transition == 'inTopSlow':
            keys = {transitionDelay:400,transitionDelay+3500:position[1]}
            c.input0 = position[0]
            animate(c,'input1',keys)
        else:
            c.input0 = position[0]
            c.input1 = position[1]

        # if we're transitioning out, die at the end of it
        if transitionOutDelay is not None:
            bs.gameTimer(transitionDelay+transitionOutDelay+1000,bs.WeakCall(self.handleMessage,bs.DieMessage()))

    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage):
            self.node.delete()
        else:
            bs.Actor.handleMessage(self,m)


class AchievementText(bsGame.Actor):
    """
    category: Game Flow Classes

    A bit of text that pops up above a position to denote something special.
    """
    def __init__(self,activity,retain,text,position=(0,0,0),color=(1,1,1,1),randomOffset=0.5,offset=(0,0,0),scale=1.0):
        """
        Instantiate with given values.

        randomOffset is the amount of random offset from the provided position that will be applied.
        This can help multiple achievements from overlapping too much.
        """
        bs.Actor.__init__(self,activity,retain)

        pos = (position[0]+offset[0]+randomOffset*(0.5-random.random()),
               position[1]+offset[0]+randomOffset*(0.5-random.random()),
               position[2]+offset[0]+randomOffset*(0.5-random.random()))
        
        self.node = bs.newNode('text',
                               owner=activity,
                               attrs={'text':text,
                                      'inWorld':True,
                                      'hAlign':'center'},
                               name='Achievement Text',
                               delegate=self)

        lifespan = 1500.0

        # scale up
        animate(self.node,'scale',{0:0.0,
                                          lifespan*0.11:0.020*0.7*scale,
                                          lifespan*0.16:0.013*0.7*scale,
                                          lifespan*0.25:0.014*0.7*scale})

        # translate upward
        self._tcombine = bs.newNode('combine',owner=self.node,attrs={'input0':pos[0],'input2':pos[2]})
        self._tcombine.connectAttr('output',self.node,'translate')
        animate(self._tcombine,'input1',{0:pos[1]+1.5,lifespan:pos[1]+2.0})

        # fade our opacity in/out
        self._combine = bs.newNode('combine',owner=self.node,attrs={'input0':color[0],'input1':color[1],'input2':color[2]})
        self._combine.connectAttr('output',self.node,'color')
        for i in range(4):
            #animate(self._combine,'input'+str(i),{0:0,0.1*lifespan:color[i],0.7*lifespan:color[i],lifespan:0})
            animate(self._combine,'input'+str(i),{0.13*lifespan:color[i],0.18*lifespan:4.0*color[i],0.22*lifespan:color[i]})
        animate(self._combine,'input3',{0:0,0.1*lifespan:color[3],
                                                #0.2*lifespan:color[3],0.24*lifespan:color[3]*3,0.28*lifespan:color[3],
                                                0.7*lifespan:color[3],lifespan:0})

        # kill ourself 
        self._dieTimer = bs.Timer(int(lifespan),bs.WeakCall(self.handleMessage,bs.DieMessage()))

    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage): self.node.delete()
        else: bs.Actor.handleMessage(self,m)
            

class TipsText(bsGame.Actor):
    """ a bit of text that shows various messages; good for helpful-tips kinda things """
    
    def __init__(self,activity,retain,offsY=100):
        bs.Actor.__init__(self,activity,retain)
        self._tipScale = 0.8
        self._tipTitleScale = 1.1
        self._offsY = offsY
        self.node = bs.newNode('text',
                               owner=activity,
                               attrs={"text":'',
                                      "scale":[self._tipScale],
                                      "hAlign":'left',
                                      "vAlign":'center',
                                      "vAttach":'bottom'},
                               name='Info Text',
                               delegate=self)
        self.titleNode = bs.newNode('text',
                               owner=activity,
                               attrs={'text':bs.getResource('tipText')+':',
                                      'scale':[self._tipTitleScale],
                                      'maxWidth':122,
                                      'hAlign':'right',
                                      'vAlign':'center',
                                      'vAttach':'bottom'},
                               name='Info Text Title',
                               delegate=self)

        self._messageDuration = 10000
        self._messageSpacing = 3000

        self._changeTimer = bs.Timer(self._messageDuration+self._messageSpacing,
                                          bs.WeakCall(self.changePhrase),repeat=True)
        self._combine = bs.newNode("combine",owner=self.node,attrs={'input0':1.0,'input1':0.8,'input2':1.0})
        self._combine.connectAttr('output',self.node,'color')
        self._combine.connectAttr('output',self.titleNode,'color')
        self.changePhrase()

    def changePhrase(self):
        nextTip = bs.translate('tips',_getNextTip())
        s = self._messageSpacing
        if self.node.exists():
            tipWidth = self._tipScale * bs.getStringWidth(nextTip)
            if tipWidth > 850: tipExtraScale = 850.0/tipWidth
            else: tipExtraScale = 1.0
            self.node.scale = [self._tipScale*tipExtraScale]
            self.node.translate = (-0.5*(tipWidth*tipExtraScale)+55+15,self._offsY)
            self.titleNode.translate = (-0.5*(tipWidth*tipExtraScale)+55,self._offsY+3)
        keys = {s:0,s+1000:1.0,s+self._messageDuration-1000:1.0,s+self._messageDuration:0.0}
        
        animate(self._combine,"input3",dict([[k,v*0.5] for k,v in keys.items()]))
        self.node.text = nextTip

    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage):
            self.node.delete()
            self.titleNode.delete()
        else: bs.Actor.handleMessage(self,m)


class ZoomText(bsGame.Actor):
    """
    category: Game Flow Classes

    Big Zooming Text Actor.
    Used for things such as the 'BOB WINS' victory messages.
    """

    def __init__(self,activity,retain,text,position=(0,0),shiftPosition=None,shiftDelay=None,lifespan=None,flash=True,trail=True,hAlign="center",
                 color=(0.9,0.4,0.0),jitter=0.0,trailColor=(1.0,0.35,0.1,0.0),scale=1.0,projectScale=1.0,tiltTranslate=0.0,maxWidth=None):
        bs.Actor.__init__(self,activity,retain)

        self._dying = False

        positionAdjusted = (position[0],position[1]-100)
        if shiftDelay == None: shiftDelay = 2500
        if shiftDelay < 0: raise Exception("negative shift delay")

        self._projectScale = projectScale
        self.node = bs.newNode("text",
                               owner=activity,
                               attrs={'translate':positionAdjusted,
                                      'big':True,
                                      'text':text,
                                      'trail':trail,
                                      'scale':[scale],
                                      'maxWidth':maxWidth if maxWidth is not None else 0.0,
                                      'tiltTranslate':tiltTranslate,
                                      'hAlign':hAlign,
                                      'vAlign':'center'},
                               name='Zoom Text: \''+text+'\'',
                               delegate=self)
            
        # if they want jitter, animate its position slightly...
        if jitter > 0.0:
            self._jitter(positionAdjusted,jitter*scale)

        # if they want shifting, move to the shift position and then resume jittering
        if shiftPosition != None:
            positionAdjusted2 = (shiftPosition[0],shiftPosition[1]-100)
            bs.gameTimer(shiftDelay,bs.WeakCall(self._shift,positionAdjusted,positionAdjusted2))
            if jitter > 0.0:
                bs.gameTimer(shiftDelay+250,bs.WeakCall(self._jitter,positionAdjusted2,jitter*scale))
                #bs.gameTimer(shiftDelay+250,(self._jitter,(positionAdjusted2,jitter)))
         

        colorCombine = bs.newNode("combine",
                                  owner=self.node,
                                  attrs={"input2":color[2],
                                         "input3":1.0})
        colorCombine.connectAttr("output",self.node,"color")

        if trail:
            trailColor = bs.newNode("combine",
                                    owner=self.node,
                                    attrs={"input0":trailColor[0],
                                           "input1":trailColor[1],
                                           "input2":trailColor[2],
                                           "input3":trailColor[3]})
            trailColor.connectAttr("output",self.node,"trailColor")
            foo = 0.85
            # animate(self.node,"trailProjectScale",{0:0, int(foo*201):0.6, int(foo*347):0.8,
            #                                                int(foo*478):0.9, int(foo*595):0.93, int(foo*748):0.95, int(foo*941):0.95})
            animate(self.node,"trailProjectScale",{0:0*projectScale, int(foo*201):0.6*projectScale, int(foo*347):0.8*projectScale,
                                                           int(foo*478):0.9*projectScale, int(foo*595):0.93*projectScale, int(foo*748):0.95*projectScale, int(foo*941):0.95*projectScale})
        if flash:
            animate(colorCombine,"input0",{0:color[0], 150:0.2+2.0*color[0]},loop=True)
            animate(colorCombine,"input1",{0:color[1], 150:0.15+2.0*color[1]},loop=True)
            animate(colorCombine,"input2",{0:color[2], 150:0.1+2.0*color[2]},loop=True)
        else:
            colorCombine.input0 = color[0]
            colorCombine.input1 = color[1]
            
        # animate(self.node,"projectScale",{0:0, 270:1.05, 300:1})
        animate(self.node,"projectScale",{0:0, 270:1.05*projectScale, 300:1*projectScale})

        # if they give us a lifespan, kill ourself down the line
        if lifespan != None:
            bs.gameTimer(lifespan,bs.WeakCall(self.handleMessage,bs.DieMessage()))

    def handleMessage(self,m):
        if isinstance(m,bs.DieMessage):
            if not self._dying and self.node.exists():
                self._dying = True
                animate(self.node,"projectScale",{0:1*self._projectScale, 600:1.2*self._projectScale})
                animate(self.node,"opacity",{0:1, 300:0})
                animate(self.node,"trailOpacity",{0:1, 600:0})
                bs.gameTimer(700,self.node.delete)
        else:
            bs.Actor.handleMessage(self,m)

    def _jitter(self,position,jitterAmount):
        if not self.node.exists(): return
        c = bs.newNode("combine",owner=self.node)
        c.connectAttr('output',self.node,'translate')

        for index,attr in enumerate(['input0','input1']):
            keys = {}
            time = 0
            # gen some random keys for that stop-motion-y look
            for i in range(10):
                keys[time] = position[index] + (random.random()-0.5)*jitterAmount*1.6
                time += random.random() * 100
            animate(c,attr,keys,loop=True)

    def _shift(self,position1,position2):
        if not self.node.exists(): return
        c = bs.newNode("combine",owner=self.node)
        c.connectAttr('output',self.node,'translate')
        animate(c,'input0',{0:position1[0],250:position2[0]})
        animate(c,'input1',{0:position1[1],250:position2[1]})


